import datetime as dt
from typing import List

from autoconf import conf
from autofit.mapper.prior_model.representative import find_groups
from autofit.text import formatter as frm, samples_text
from autofit.tools.util import info_whitespace


def padding(item, target=6):
    string = str(item)
    difference = target - len(string)
    prefix = difference * " "
    return f"{prefix}{string}"


def result_max_lh_info_from(max_log_likelihood_sample : List[float], max_log_likelihood : float, model) -> List[str]:
    """
    Output the maximum log likelihood model only, for quick reference.
    """
    results = []

    results += [
        frm.add_whitespace(
            str0="Maximum Log Likelihood ",
            str1="{:.8f}".format(max_log_likelihood),
            whitespace=info_whitespace(),
        )
    ]

    results += ["\n\n", model.parameterization]

    results += ["\n\nMaximum Log Likelihood Model:\n\n"]

    formatter = frm.TextFormatter(line_length=info_whitespace())

    paths = []

    for (_, prior), value in zip(
            model.unique_path_prior_tuples,
            max_log_likelihood_sample,
    ):
        for path in model.all_paths_for_prior(prior):
            paths.append((path, value))

    for path, value in find_groups(paths):
        formatter.add(path, format_str().format(value))
    results += [formatter.text + "\n"]

    return results


def result_info_from(samples) -> str:
    """
    Output the full model.results file, which include the most-likely model, most-probable model at 1 and 3
    sigma confidence and information on the maximum log likelihood.
    """
    results = []

    if hasattr(samples, "log_evidence"):
        if samples.log_evidence is not None:
            results += [
                frm.add_whitespace(
                    str0="Bayesian Evidence ",
                    str1="{:.8f}".format(samples.log_evidence),
                    whitespace=info_whitespace(),
                )
            ]
            results += ["\n"]

    max_log_likelihood_sample = samples.max_log_likelihood(as_instance=False)

    results += result_max_lh_info_from(
        max_log_likelihood_sample=max_log_likelihood_sample,
        max_log_likelihood=(max(samples.log_likelihood_list)),
        model=samples.model,
    )

    if hasattr(samples, "pdf_converged"):
        if samples.pdf_converged:
            results += samples_text.summary(
                samples=samples, sigma=3.0, indent=4, line_length=info_whitespace()
            )
            results += ["\n"]
            results += samples_text.summary(
                samples=samples, sigma=1.0, indent=4, line_length=info_whitespace()
            )

        else:
            results += [
                "\n WARNING: The samples have not converged enough to compute a PDF and model errors. \n "
                "The model below over estimates errors. \n\n"
            ]
            results += samples_text.summary(
                samples=samples, sigma=1.0, indent=4, line_length=info_whitespace()
            )

        results += ["\n\ninstances\n"]

    formatter = frm.TextFormatter(line_length=info_whitespace())

    for path, value in find_groups(samples.model.path_float_tuples):
        formatter.add(path, value)

    results += ["\n" + formatter.text]

    return "".join(results)


def search_summary_from_samples(samples) -> [str]:
    line = [f"Total Samples = {samples.total_samples}\n"]
    if hasattr(samples, "total_accepted_samples"):
        line.append(f"Total Accepted Samples = {samples.total_accepted_samples}\n")
        line.append(f"Acceptance Ratio = {samples.acceptance_ratio}\n")
    if samples.time is not None:
        line.append(f"Time To Run = {dt.timedelta(seconds=float(samples.time))}\n")
        line.append(
            f"Time Per Sample (seconds) = {float(samples.time) / samples.total_samples}\n"
        )
    return line


def search_summary_to_file(
        samples,
        log_likelihood_function_time,
        filename,
        visualization_time=None,
):
    summary = search_summary_from_samples(samples=samples)
    summary.append(
        f"Log Likelihood Function Evaluation Time (seconds) = {log_likelihood_function_time}\n"
    )

    expected_time = dt.timedelta(
        seconds=float(samples.total_samples * log_likelihood_function_time)
    )
    summary.append(f"Expected Time To Run (seconds) = {expected_time}\n")

    try:
        speed_up_factor = float(expected_time.total_seconds()) / float(samples.time)
        summary.append(
            f"Speed Up Factor (e.g. due to parallelization) = {speed_up_factor}\n"
        )
    except TypeError:
        pass

    if visualization_time is not None:
        summary.append(
            f"Visualization Time (seconds) = {visualization_time}"
        )

    frm.output_list_of_strings_to_file(file=filename, list_of_strings=summary)


def format_str() -> str:
    """The format string for the model.results file, describing to how many decimal points every parameter
    estimate is output in the model.results file.
    """
    decimal_places = conf.instance["general"]["output"]["model_results_decimal_places"]
    return f"{{:.{decimal_places}f}}"



import numpy as np

from autoconf import cached_property
from autofit.graphical.utils import (
    nested_filter,
    nested_update,
    is_variable,
    to_variabledata,
)
from autofit.mapper.variable import (
    Variable,
    FactorValue,
    VariableData,
    VariableLinearOperator,
)
from autofit.mapper.variable_operator import (
    RectVariableOperator,
)
from abc import ABC
from typing import (
    Tuple,
    Dict,
    Union,
    Callable,
)

Protocol = ABC  # for python 3.7 compat

Value = Dict[Variable, np.ndarray]
GradientValue = VariableData


class FactorInterface(Protocol):
    def __call__(self, values: Value) -> FactorValue:
        pass


class FactorGradientInterface(Protocol):
    def __call__(self, values: Value) -> Tuple[FactorValue, GradientValue]:
        pass


class AbstractJacobian(VariableLinearOperator):
    """
    Examples
    --------
    def linear(x, a, b):
        z = x.dot(a) + b
        return (z**2).sum(), z

    def full(x, a, b):
        z2, z = linear(x, a, b)
        return z2 + z.sum()

    x_, a_, b_, y_, z_ = variables("x, a, b, y, z")
    x = np.arange(10.).reshape(5, 2)
    a = np.arange(2.).reshape(2, 1)
    b = np.ones(1)
    y = np.arange(0., 10., 2).reshape(5, 1)
    # values = {x_: x, y_: y, a_: a, b_: b}

    linear_factor_jvp = FactorJVP(
        linear, x_, a_, b_, factor_out=(FactorValue, z_))

    linear_factor_vjp = FactorVJP(
        linear, x_, a_, b_, factor_out=(FactorValue, z_))

    values = {x_: x, a_: a, b_: b}

    jvp_val, jvp_jac = linear_factor_jvp.func_jacobian(values)
    vjp_val, vjp_jac = linear_factor_vjp.func_jacobian(values)


    assert np.allclose(vjp_val, jvp_val)
    assert (vjp_jac(vjp_val) - jvp_jac(vjp_val)).norm() == 0
    """

    def __call__(self, values):
        return self.__rmul__(values)

    def __str__(self) -> str:
        out_var = str(
            nested_update(self.factor_out, {v: v.name for v in self.out_variables})
        ).replace("'", "")

        in_var = ", ".join(v.name for v in self.variables)
        cls_name = type(self).__name__
        return f"{cls_name}({out_var} → ∂({in_var})ᵀ {out_var})"

    __repr__ = __str__

    def _full_repr(self) -> str:
        out_var = str(self.factor_out)
        in_var = str(self.variables)
        cls_name = type(self).__name__
        return f"{cls_name}({out_var} → ∂({in_var})ᵀ {out_var})"

    def grad(self, values=None):
        grad = VariableData({FactorValue: 1.0})
        if values:
            grad.update(values)

        jac = self(grad)
        for v, g in jac.items():
            grad[v] = grad.get(v, 0) + g

        grad.pop(FactorValue)
        return grad


class JacobianVectorProduct(AbstractJacobian, RectVariableOperator):
    __init__ = RectVariableOperator.__init__

    @property
    def variables(self):
        return self.left_variables

    @property
    def out_variables(self):
        return self.right_variables

    @property
    def factor_out(self):
        return tuple(self.out_variables)


class VectorJacobianProduct(AbstractJacobian):
    def __init__(
        self,
        factor_out,
        vjp: Callable,
        *args: Variable,
        out_shapes=None,
    ):
        self.factor_out = factor_out
        self.vjp = vjp
        self._args = args
        self._variables = tuple(v for v, in nested_filter(is_variable, args))
        self.out_shapes = out_shapes

    @property
    def args(self):
        return self._args

    @property
    def variables(self):
        return self._variables

    @cached_property
    def out_variables(self):
        return set(v[0] for v in nested_filter(is_variable, self.factor_out))

    def _get_cotangent(self, values):
        if isinstance(values, FactorValue):
            values = values.to_dict()

        if isinstance(values, dict):
            if self.out_shapes:
                for v in self.out_shapes.keys() - values.keys():
                    values[v] = np.zeros(self.out_shapes[v])
            out = nested_update(self.factor_out, values)
            return out

        if isinstance(values, int):
            values = float(values)

        return values

    def __call__(self, values: Union[VariableData, FactorValue]) -> VariableData:
        v = self._get_cotangent(values)
        grads = self.vjp(v)
        return to_variabledata(self.args, grads)

    __rmul__ = __call__

    def _not_implemented(self, *args):
        raise NotImplementedError()

    __rtruediv__ = _not_implemented
    ldiv = _not_implemented
    __mul__ = _not_implemented
    update = _not_implemented



import os
from typing import Dict, Optional

from autofit.database.sqlalchemy_ import sa

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.search.nest import abstract_nest
from autofit.non_linear.fitness import Fitness
from autofit.non_linear.samples.sample import Sample
from autofit.non_linear.samples.nest import SamplesNest

class UltraNest(abstract_nest.AbstractNest):
    __identifier_fields__ = (
        "draw_multiple",
        "ndraw_min",
        "ndraw_max",
        "min_num_live_points",
        "cluster_num_live_points",
        "insertion_test_zscore_threshold",
        "stepsampler_cls",
        "nsteps"
    )

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            iterations_per_quick_update: int = None,
            iterations_per_full_update: int = None,
            number_of_cores: int = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        """
        An UltraNest non-linear search.

        UltraNest is an optional requirement and must be installed manually via the command `pip install ultranest`.
        It is optional as it has certain dependencies which are generally straight forward to install (e.g. Cython).

        For a full description of UltraNest and its Python wrapper PyUltraNest, checkout its Github and documentation
        webpages:

        https://github.com/JohannesBuchner/UltraNest
        https://johannesbuchner.github.io/UltraNest/readme.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        iterations_per_full_update
            The number of iterations performed between update (e.g. output latest model to hard-disk, visualization).
        number_of_cores
            The number of cores sampling is performed using a Python multiprocessing Pool instance.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            iterations_per_quick_update=iterations_per_quick_update,
            iterations_per_full_update=iterations_per_full_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

        for key, value in self.config_dict_stepsampler.items():
            setattr(self, key, value)
            if self.config_dict_stepsampler["stepsampler_cls"] is None:
                self.nsteps = None

        self.logger.debug("Creating UltraNest Search")

    def _fit(self, model: AbstractPriorModel, analysis):
        """
        Fit a model using the search and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that includes the maximum log likelihood instance and full
        set of accepted ssamples of the fit.
        """

        try:
            import ultranest
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "\n--------------------\n"
                "You are attempting to perform a model-fit using UltraNest. \n\n"
                "However, the optional library UltraNest (https://johannesbuchner.github.io/UltraNest/index.html) is "
                "not installed.\n\n"
                "Install it via the command `pip install ultranest==3.6.2`.\n\n"
                "----------------------"
            )

        fitness = Fitness(
            model=model,
            analysis=analysis,
            paths=self.paths,
            fom_is_log_likelihood=True,
            resample_figure_of_merit=-1.0e99
        )

        def prior_transform(cube):
            return model.vector_from_unit_vector(
                unit_vector=cube,
            )

        log_dir = self.paths.search_internal_path

        try:
            checkpoint_exists = os.path.exists(log_dir / "chains")
        except TypeError:
            checkpoint_exists = False

        if checkpoint_exists:
            self.logger.info(
                "Resuming UltraNest non-linear search (previous samples found)."
            )
        else:
            self.logger.info(
                "Starting new UltraNest non-linear search (no previous samples found)."
            )

        search_internal = ultranest.ReactiveNestedSampler(
            param_names=model.parameter_names,
            loglike=fitness.call_wrap,
            transform=prior_transform,
            log_dir=log_dir,
            **self.config_dict_search
        )

        search_internal.stepsampler = self.stepsampler

        finished = False

        while not finished:

            try:
                total_iterations = search_internal.ncall
            except AttributeError:
                total_iterations = 0

            if self.config_dict_run["max_ncalls"] is not None:
                iterations = self.config_dict_run["max_ncalls"]
            else:
                iterations = total_iterations + self.iterations_per_full_update

            if iterations > 0:

                filter_list = ["max_ncalls", "dkl", "lepsilon"]
                config_dict_run = {
                    key: value for key, value
                    in self.config_dict_run.items()
                    if key
                    not in filter_list
                }

                config_dict_run["update_interval_ncall"] = iterations

                search_internal.run(
                    max_ncalls=iterations,
                    **config_dict_run
                )

            self.paths.save_search_internal(
                  obj=search_internal.results,
              )

            iterations_after_run = search_internal.ncall

            if (
                    total_iterations == iterations_after_run
                    or iterations_after_run == self.config_dict_run["max_ncalls"]
            ):
                finished = True

            if not finished:

                self.perform_update(
                    model=model,
                    analysis=analysis,
                    during_analysis=True,
                    fitness=fitness,
                    search_internal=search_internal
                )

        return search_internal, fitness

    def output_search_internal(self, search_internal):
        """
        Output the sampler results to hard-disk in their internal format.

        UltraNest uses a backend to store and load results, therefore the outputting of the search internal to a
        dill file is disabled.

        However, a dictionary of the search results is output to dill above.

        Parameters
        ----------
        sampler
            The nautilus sampler object containing the results of the model-fit.
        """
        pass

    def samples_info_from(self, search_internal=None):

        search_internal = search_internal or self.paths.load_search_internal()

        return {
            "log_evidence": search_internal["logz"],
            "total_samples": search_internal["ncall"],
            "total_accepted_samples": len(search_internal["weighted_samples"]["logl"]),
            "time": self.timer.time if self.timer else None,
            "number_live_points": self.config_dict_run["min_num_live_points"]
        }

    def samples_via_internal_from(self, model: AbstractPriorModel, search_internal=None):
        """
        Returns a `Samples` object from the ultranest internal results.

        The samples contain all information on the parameter space sampling (e.g. the parameters,
        log likelihoods, etc.).

        The internal search results are converted from the native format used by the search to lists of values
        (e.g. `parameter_lists`, `log_likelihood_list`).

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        """

        search_internal = search_internal.results or self.paths.load_search_internal()

        parameters = search_internal["weighted_samples"]["points"]
        log_likelihood_list = search_internal["weighted_samples"]["logl"]
        log_prior_list = [
            sum(model.log_prior_list_from_vector(vector=vector)) for vector in parameters
        ]
        weight_list = search_internal["weighted_samples"]["weights"]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list
        )

        return SamplesNest(
            model=model,
            sample_list=sample_list,
            samples_info=self.samples_info_from(search_internal=search_internal),
        )

    def config_dict_test_mode_from(self, config_dict: Dict) -> Dict:
        """
        Returns a configuration dictionary for test mode meaning that the sampler terminates as quickly as possible.

        Entries which set the total number of samples of the sampler (e.g. maximum calls, maximum likelihood
        evaluations) are reduced to low values meaning it terminates nearly immediately.

        Parameters
        ----------
        config_dict
            The original configuration dictionary for this sampler which includes entries controlling how fast the
            sampler terminates.

        Returns
        -------
        A configuration dictionary where settings which control the sampler's number of samples are reduced so it
        terminates as quickly as possible.
        """
        return {
            **config_dict,
            "max_iters": 1,
            "max_ncalls": 1,
        }

    @property
    def config_dict_stepsampler(self):

        config_dict = {}

        config_dict_step = self.config_type[self.__class__.__name__]["stepsampler"]

        for key, value in config_dict_step.items():
            try:
                config_dict[key] = self.kwargs[key]
            except KeyError:
                config_dict[key] = value

        return config_dict

    @property
    def stepsampler(self):

        from ultranest import stepsampler

        config_dict_stepsampler = self.config_dict_stepsampler
        stepsampler_cls = config_dict_stepsampler["stepsampler_cls"]
        config_dict_stepsampler.pop("stepsampler_cls")

        if stepsampler_cls is None:
            return None
        elif stepsampler_cls == "RegionMHSampler":
            return stepsampler.RegionMHSampler(**config_dict_stepsampler)
        elif stepsampler_cls == "AHARMSampler":
            config_dict_stepsampler.pop("scale")
            return stepsampler.AHARMSampler(**config_dict_stepsampler)
        elif stepsampler_cls == "CubeMHSampler":
            return stepsampler.CubeMHSampler(**config_dict_stepsampler)
        elif stepsampler_cls == "CubeSliceSampler":
            return stepsampler.CubeSliceSampler(**config_dict_stepsampler)
        elif stepsampler_cls == "RegionSliceSampler":
            return stepsampler.RegionSliceSampler(**config_dict_stepsampler)


from autofit.non_linear.samples import SamplesPDF, Sample, SamplesNest


def samples_with_log_likelihood_list(log_likelihood_list):
    return [
        Sample(log_likelihood=log_likelihood, log_prior=0, weight=0)
        for log_likelihood in log_likelihood_list
    ]


class MockSamples(SamplesPDF):
    def __init__(
        self,
        model=None,
        sample_list=None,
        samples_info=None,
        log_likelihood_list=None,
        prior_means=None,
        **kwargs,
    ):
        self._log_likelihood_list = log_likelihood_list

        self.model = model

        sample_list = sample_list or self.default_sample_list

        samples_info = samples_info or {"unconverged_sample_size": 0}

        super().__init__(
            model=model,
            sample_list=sample_list,
            samples_info=samples_info,
            **kwargs,
        )

    @property
    def default_sample_list(self):
        return [
            Sample(
                log_likelihood=log_likelihood,
                log_prior=0.0,
                weight=0.0,
                kwargs={path: 1.0 for path in self.model.paths} if self.model else {},
            )
            for log_likelihood in range(3)
        ]

    @property
    def log_likelihood_list(self):
        if self._log_likelihood_list is None:
            return super().log_likelihood_list

        return self._log_likelihood_list

    @property
    def unconverged_sample_size(self):
        return self.samples_info["unconverged_sample_size"]


class MockSamplesNest(SamplesNest):
    def __init__(
        self,
        model,
        sample_list=None,
        samples_info=None,
    ):
        self.model = model

        if sample_list is None:
            sample_list = [
                Sample(log_likelihood=log_likelihood, log_prior=0.0, weight=0.0)
                for log_likelihood in self.log_likelihood_list
            ]

        super().__init__(
            model=model, sample_list=sample_list, samples_info=samples_info
        )



from typing import Dict, Optional

import numpy as np

from autofit import exc
from autofit.database.sqlalchemy_ import sa
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.fitness import Fitness
from autofit.non_linear.initializer import AbstractInitializer
from autofit.non_linear.search.mle.abstract_mle import AbstractMLE
from autofit.non_linear.samples.sample import Sample
from autofit.non_linear.samples.samples import Samples


class FitnessPySwarms(Fitness):
    def __call__(self, parameters, *kwargs):
        """
        Interfaces with any non-linear in order to fit a model to the data and return a log likelihood via
        an `Analysis` class.

        The interface is described in full in the `__init__` docstring.

        `PySwarms` have a unique interface in that lists of parameters corresponding to multiple particles are
        passed to the fitness function. A bespoke `__call__` method is therefore required to handle this.

        The figure of merit is the log posterior multiplied by -2.0, which is the chi-squared value that is minimized
        by the `PySwarms` non-linear search.

        Parameters
        ----------
        parameters
            The parameters (typically a list) chosen by a non-linear search, which are mapped to an instance of the
            model via its priors and fitted to the data.
        kwargs
            Addition key-word arguments that may be necessary for specific non-linear searches.

        Returns
        -------
        The figure of merit returned to the non-linear search, which is either the log likelihood or log posterior.
        """

        if isinstance(parameters[0], float):
            parameters = [parameters]

        figure_of_merit_list = []

        for params_of_particle in parameters:
            try:
                instance = self.model.instance_from_vector(vector=params_of_particle)
                log_likelihood = self.analysis.log_likelihood_function(
                    instance=instance
                )
                log_prior = self.model.log_prior_list_from_vector(
                    vector=params_of_particle
                )
                log_posterior = log_likelihood + sum(log_prior)
                figure_of_merit = -2.0 * log_posterior
            except exc.FitException:
                figure_of_merit = np.nan

            if np.isnan(figure_of_merit):
                figure_of_merit = -2.0 * self.resample_figure_of_merit

            figure_of_merit_list.append(figure_of_merit)

        return np.asarray(figure_of_merit_list)


class AbstractPySwarms(AbstractMLE):
    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        initializer: Optional[AbstractInitializer] = None,
        iterations_per_quick_update: int = None,
        iterations_per_full_update: int = None,
        number_of_cores: int = None,
        session: Optional[sa.orm.Session] = None,
        **kwargs
    ):
        """
        A PySwarms Particle Swarm MLE global non-linear search.

        For a full description of PySwarms, checkout its Github and readthedocs webpages:

        https://github.com/ljvmiranda921/pyswarms
        https://pyswarms.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        number_of_cores
            The number of cores sampling is performed using a Python multiprocessing Pool instance.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            initializer=initializer,
            iterations_per_quick_update=iterations_per_quick_update,
            iterations_per_full_update=iterations_per_full_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

        self.logger.debug("Creating PySwarms Search")

    def _fit(self, model: AbstractPriorModel, analysis):
        """
        Fit a model using PySwarms and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """

        fitness = FitnessPySwarms(
            model=model,
            analysis=analysis,
            fom_is_log_likelihood=False,
            resample_figure_of_merit=-np.inf,
            convert_to_chi_squared=True,
        )

        try:
            search_internal = self.paths.load_search_internal()

            init_pos = search_internal.pos_history[-1]
            total_iterations = len(search_internal.cost_history)

            self.logger.info(
                "Resuming PySwarms non-linear search (previous samples found)."
            )

        except (FileNotFoundError, TypeError, AttributeError):
            (
                unit_parameter_lists,
                parameter_lists,
                log_posterior_list,
            ) = self.initializer.samples_from_model(
                total_points=self.config_dict_search["n_particles"],
                model=model,
                fitness=fitness,
                paths=self.paths,
                n_cores=self.number_of_cores,
            )

            init_pos = np.zeros(
                shape=(self.config_dict_search["n_particles"], model.prior_count)
            )

            for index, parameters in enumerate(parameter_lists):
                init_pos[index, :] = np.asarray(parameters)

            total_iterations = 0

            self.logger.info(
                "Starting new PySwarms non-linear search (no previous samples found)."
            )

        ## TODO : Use actual limits

        vector_lower = model.vector_from_unit_vector(
            unit_vector=[1e-6] * model.prior_count,
        )
        vector_upper = model.vector_from_unit_vector(
            unit_vector=[0.9999999] * model.prior_count,
        )

        lower_bounds = [lower for lower in vector_lower]
        upper_bounds = [upper for upper in vector_upper]

        bounds = (np.asarray(lower_bounds), np.asarray(upper_bounds))

        while total_iterations < self.config_dict_run["iters"]:
            search_internal = self.search_internal_from(
                model=model, fitness=fitness, bounds=bounds, init_pos=init_pos
            )

            iterations_remaining = self.config_dict_run["iters"] - total_iterations

            iterations = min(self.iterations_per_full_update, iterations_remaining)

            if iterations > 0:
                search_internal.optimize(
                    objective_func=fitness.call_wrap, iters=int(iterations)
                )

                total_iterations += iterations

                # TODO : Running PySwarms in NoteBook raises
                # TODO: TypeError: cannot pickle '_hashlib.HMAC' object

                self.output_search_internal(search_internal=search_internal)

                self.perform_update(
                    model=model,
                    analysis=analysis,
                    during_analysis=True,
                    fitness=fitness,
                    search_internal=search_internal,
                )

                init_pos = search_internal.pos_history[-1]

        return search_internal, fitness

    def output_search_internal(self, search_internal):
        try:
            self.paths.save_search_internal(
                obj=search_internal,
            )
        except TypeError:
            pass

    def samples_via_internal_from(self, model, search_internal=None):
        """
        Returns a `Samples` object from the pyswarms internal results.

        The samples contain all information on the parameter space sampling (e.g. the parameters,
        log likelihoods, etc.).

        The internal search results are converted from the native format used by the search to lists of values
        (e.g. `parameter_lists`, `log_likelihood_list`).

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        """

        if search_internal is None:
            search_internal = self.paths.load_search_internal()

        search_internal_dict = {
            "total_iterations": None,
            "log_posterior_list": [
                -0.5 * cost for cost in search_internal.cost_history
            ],
            "time": self.timer.time if self.timer else None,
        }
        pos_history = search_internal.pos_history

        parameter_lists = [
            param.tolist() for parameters in pos_history for param in parameters
        ]
        parameter_lists_2 = [parameters.tolist()[0] for parameters in pos_history]

        log_posterior_list = search_internal_dict["log_posterior_list"]
        log_prior_list = model.log_prior_list_from(parameter_lists=parameter_lists)
        log_likelihood_list = [
            lp - prior for lp, prior in zip(log_posterior_list, log_prior_list)
        ]
        weight_list = len(log_likelihood_list) * [1.0]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists_2,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list,
        )

        return Samples(
            model=model,
            sample_list=sample_list,
            samples_info=search_internal_dict,
        )

    def config_dict_test_mode_from(self, config_dict: Dict) -> Dict:
        """
        Returns a configuration dictionary for test mode meaning that the sampler terminates as quickly as possible.

        Entries which set the total number of samples of the sampler (e.g. maximum calls, maximum likelihood
        evaluations) are reduced to low values meaning it terminates nearly immediately.

        Parameters
        ----------
        config_dict
            The original configuration dictionary for this sampler which includes entries controlling how fast the
            sampler terminates.

        Returns
        -------
        A configuration dictionary where settings which control the sampler's number of samples are reduced so it
        terminates as quickly as possible.
        """
        return {
            **config_dict,
            "iters": 1,
        }

    def search_internal_from(self, model, fitness, bounds, init_pos):
        raise NotImplementedError()



from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Callable


class AbstractColumn(ABC):
    def __init__(self, name: str):
        """
        A column in the summary table.

        Parameters
        ----------
        name
            An optional name for the column
        """
        self.name = name

    @abstractmethod
    def value(self, row: "Row"):
        pass


class ValueType(Enum):
    """
    Possible value types to create columns for.
    """

    Median = 0  # The median_pdf_sample value
    MaxLogLikelihood = 1  # The max likelihood value
    ValuesAt1Sigma = 2  # The values at 1 sigma. This includes a lower and upper value
    ValuesAt3Sigma = 3  # The values at 3 sigma. This includes a lower and upper value


class Column(AbstractColumn):
    def __init__(
        self,
        argument: str,
        name: Optional[str] = None,
        value_types: list[ValueType] = (ValueType.Median,),
    ):
        """
        A column in the summary table.

        Parameters
        ----------
        argument
            The argument as it appears in the median_pdf_sample arguments
        name
            An optional name for the column
        value_types
            Value types to include. See ValueType.
        """
        super().__init__(
            name
            or argument.replace(
                ".",
                "_",
            )
        )
        self.argument = argument
        self.value_types = value_types

    def value(self, row: "Row"):
        result = {}

        if ValueType.Median in self.value_types:
            try:
                result[""] = row.median_pdf_sample_kwargs[self.path]
            except KeyError:
                result[""] = None

        if ValueType.MaxLogLikelihood in self.value_types:
            try:
                result["max_lh"] = row.max_likelihood_kwargs[self.path]
            except KeyError:
                result["max_lh"] = None

        if ValueType.ValuesAt1Sigma in self.value_types:
            try:
                lower, upper = row.values_at_sigma_1_kwargs[self.path]
                result["lower_1_sigma"] = lower
                result["upper_1_sigma"] = upper
            except KeyError:
                result["lower_1_sigma"] = None
                result["upper_1_sigma"] = None

        if ValueType.ValuesAt3Sigma in self.value_types:
            try:
                lower, upper = row.values_at_sigma_3_kwargs[self.path]
                result["lower_3_sigma"] = lower
                result["upper_3_sigma"] = upper
            except KeyError:
                result["lower_3_sigma"] = None
                result["upper_3_sigma"] = None

        return result

    @property
    def path(self) -> tuple:
        """
        The path of an argument in the median_pdf_sample arguments.
        """
        return tuple(self.argument.split("."))


class ComputedColumn(AbstractColumn):
    def __init__(self, name: str, compute: Callable):
        """
        A column in the summary table that is computed from other columns.

        Parameters
        ----------
        name
            The name of the column
        compute
            A function that takes the median_pdf_sample arguments and returns the computed value
        """
        super().__init__(name)
        self.compute = compute

    def value(self, row: "Row"):

        try:
            return self.compute(row.result)
        except AttributeError as e:
            raise AssertionError(
                "Cannot compute additional fields if no samples.json present"
            ) from e


class LabelColumn(AbstractColumn):
    def __init__(self, name: str, values: list):
        """
        A column in the summary table that is a label.

        Parameters
        ----------
        name
            The name of the column
        """
        super().__init__(name)
        self.values = values

    def value(self, row: "Row"):
        return self.values[row.number]



import logging
from functools import wraps
from sqlalchemy import text
from typing import Optional
from ..sqlalchemy_ import sa

logger = logging.getLogger(__name__)


def needs_revision_table(func):
    """
    Applies to functions that depend on the existence
    of the revision table. If the table does not exist
    it is created and then the function is executed.

    If the table already existed but an OperationalError
    is raised then that error is propagated.

    Parameters
    ----------
    func
        Some function that depends on the revision table

    Returns
    -------
    A decorated function
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except sa.exc.OperationalError as e:
            if self.is_table:
                raise e
            self._init_revision_table()
            return func(self, *args, **kwargs)

    return wrapper


class SessionWrapper:
    def __init__(self, session: sa.orm.Session):
        """
        Wraps a SQLAlchemy session so that certain commands can be
        encapsulated.

        Parameters
        ----------
        session
        """
        self.session = session

    def _init_revision_table(self):
        """
        Creates the revision table with a single null entry
        """
        self.session.execute(
            text("CREATE TABLE revision (revision_id VARCHAR PRIMARY KEY)")
        )
        self.session.execute(text("INSERT INTO revision (revision_id) VALUES (null)"))

    @property
    def is_table(self) -> bool:
        """
        Does the revision table exist?
        """
        try:
            self.session.execute(text("SELECT 1 FROM revision"))
            return True
        except sa.exc.OperationalError:
            return False

    @property
    @needs_revision_table
    def revision_id(self) -> Optional[str]:
        """
        Describes the current revision of the database. None if no
        revisions have been made.
        """
        for row in self.session.execute(text("SELECT revision_id FROM revision")):
            return row[0]
        return None

    @revision_id.setter
    @needs_revision_table
    def revision_id(self, revision_id: str):
        self.session.execute(
            text(f"UPDATE revision SET revision_id = :revision_id"),
            {"revision_id": revision_id},
        )



from __future__ import annotations

from pathlib import Path
from typing import Optional, Union


from autofit.database.sqlalchemy_ import sa

from autofit.mapper.prior_model.abstract import AbstractPriorModel

from .abstract import AbstractDynesty, prior_transform

class DynestyStatic(AbstractDynesty):
    __identifier_fields__ = (
        "nlive",
        "bound",
        "sample",
        "bootstrap",
        "enlarge",
        "walks",
        "facc",
        "slices",
        "fmove",
        "max_move",
    )

    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[Union[str, Path]] = None,
        unique_tag: Optional[str] = None,
        iterations_per_quick_update: int = None,
        iterations_per_full_update: int = None,
        number_of_cores: int = None,
        session: Optional[sa.orm.Session] = None,
        **kwargs,
    ):
        """
        A Dynesty `NonLinearSearch` using a static number of live points.

        For a full description of Dynesty, checkout its GitHub and readthedocs webpages:

        https://github.com/joshspeagle/dynesty
        https://dynesty.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        iterations_per_full_update
            The number of iterations performed between update (e.g. output latest model to hard-disk, visualization).
        number_of_cores
            The number of cores sampling is performed using a Python multiprocessing Pool instance.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            iterations_per_quick_update=iterations_per_quick_update,
            iterations_per_full_update=iterations_per_full_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs,
        )

    @property
    def search_internal(self):
        from dynesty import NestedSampler as StaticSampler
        return StaticSampler.restore(self.checkpoint_file)

    def search_internal_from(
        self,
        model: AbstractPriorModel,
        fitness,
        checkpoint_exists: bool,
        pool: Optional,
        queue_size: Optional[int],
    ):
        """
        Returns an instance of the Dynesty static sampler set up using the input variables of this class.

        If no existing dynesty sampler exist on hard-disk (located via a `checkpoint_file`) a new instance is
        created with which sampler is performed. If one does exist, the dynesty `restore()` function is used to
        create the instance of the sampler.

        Dynesty samplers with a multiprocessing pool may be created by inputting a dynesty `Pool` object, however
        non pooled instances can also be created by passing `pool=None` and `queue_size=None`.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space.
        fitness
            An instance of the fitness class used to evaluate the likelihood of each model.
        pool
            A dynesty Pool object which performs likelihood evaluations over multiple CPUs.
        queue_size
            The number of CPU's over which multiprocessing is performed, determining how many samples are stored
            in the dynesty queue for samples.
        """
        from dynesty import NestedSampler as StaticSampler

        if checkpoint_exists:
            search_internal = StaticSampler.restore(
                fname=self.checkpoint_file, pool=pool
            )

            uses_pool = self.read_uses_pool()

            self.check_pool(uses_pool=uses_pool, pool=pool)

            return search_internal

        else:
            live_points = self.live_points_init_from(model=model, fitness=fitness)

            if pool is not None:
                self.write_uses_pool(uses_pool=True)
                return StaticSampler(
                    loglikelihood=pool.loglike,
                    prior_transform=pool.prior_transform,
                    ndim=model.prior_count,
                    live_points=live_points,
                    queue_size=queue_size,
                    pool=pool,
                    **self.config_dict_search,
                )

            self.write_uses_pool(uses_pool=False)
            return StaticSampler(
                loglikelihood=fitness,
                prior_transform=prior_transform,
                ndim=model.prior_count,
                logl_args=[model, fitness],
                ptform_args=[model],
                live_points=live_points,
                **self.config_dict_search,
            )

    @property
    def number_live_points(self):
        return self.config_dict_search["nlive"]



import math
import warnings
from typing import Tuple, Union

import numpy as np

from autoconf import cached_property
from ..messages.abstract import AbstractMessage


def grad_betaln(ab: np.ndarray) -> np.ndarray:
    """
    Compute the gradient of the log Beta function with respect to parameters a and b.

    Parameters
    ----------
    ab
        Array of shape (N, 2) where each row contains parameters [a, b] of Beta distributions.

    Returns
    -------
    Gradient array of shape (N, 2) with derivatives of log Beta function w.r.t a and b.
    """
    from scipy.special import psi

    psiab = psi(ab.sum(axis=1, keepdims=True))
    return psi(ab) - psiab


def jac_grad_betaln(ab: np.ndarray) -> np.ndarray:
    """
    Compute the Jacobian matrix of the gradient of the log Beta function.

    Parameters
    ----------
    ab
        Array of shape (N, 2) with Beta parameters [a, b].

    Returns
    -------
    Array of shape (N, 2, 2), the Jacobian matrices for each parameter pair.
    """
    from scipy.special import polygamma

    psi1ab = polygamma(1, ab.sum(axis=1, keepdims=True))
    fii = polygamma(1, ab) - psi1ab
    fij = -psi1ab[:, 0]
    return np.array([[fii[:, 0], fij], [fij, fii[:, 1]]]).T


def inv_beta_suffstats(
    lnX: Union[np.ndarray, float],
    ln1X: Union[np.ndarray, float],
) -> Tuple[Union[np.ndarray, float], Union[np.ndarray, float]]:
    """
    Solve for Beta distribution parameters (a, b) given log sufficient statistics.

    The system solves:

        psi(a) - psi(a + b) = lnX
        psi(b) - psi(a + b) = ln1X

    Parameters
    ----------
    lnX
        Logarithm of the expected value of X.
    ln1X
        Logarithm of the expected value of 1 - X.

    Returns
    -------
    a
        Estimated alpha parameter(s) of the Beta distribution.
    b
        Estimated beta parameter(s) of the Beta distribution.

    Warnings
    --------
    Emits a RuntimeWarning if negative parameters are found, and clamps them to 0.5.
    """

    _lnX, _ln1X = np.ravel(lnX), np.ravel(ln1X)
    lnXs = np.c_[_lnX, _ln1X]

    # Find initial starting location
    Gs = np.exp(lnXs)
    dG = 1 - Gs.sum(axis=1, keepdims=True)
    ab = np.maximum(1, (1 + Gs / dG) / 2)

    # 5 Newton Raphson itertions is generally enough
    for i in range(5):
        f = grad_betaln(ab) - lnXs
        jac = jac_grad_betaln(ab)
        ab += np.linalg.solve(jac, - f)

    if np.any(ab < 0):
        warnings.warn(
            "invalid negative parameters found for inv_beta_suffstats, "
            "clampling value to 0.5",
            RuntimeWarning
        )
        b = np.clip(ab, 0.5, None)

    shape = np.shape(lnX)
    if shape:
        a = ab[:, 0].reshape(shape)
        b = ab[:, 1].reshape(shape)
    else:
        a, b = ab[0, :]

    return a, b


class BetaMessage(AbstractMessage):

    log_base_measure = 0
    _support = ((0, 1),)
    _min = 0
    _max = 1
    _range = 1
    _parameter_support = ((0, np.inf), (0, np.inf))

    def __init__(
        self,
        alpha: float = 0.5,
        beta: float = 0.5,
        log_norm: float = 0,
        id_: Union[str, None] = None,
    ):
        """
        Represents a Beta distribution message in natural parameter form.

        Parameters
        ----------
        alpha
            Alpha (shape) parameter of the Beta distribution. Default is 0.5.
        beta
            Beta (shape) parameter of the Beta distribution. Default is 0.5.
        log_norm
            Logarithm of normalization constant for message passing. Default is 0.
        id_
            Identifier for the message. Default is None.
        """
        self.alpha = alpha
        self.beta = beta
        super().__init__(
            alpha,
            beta,
            log_norm=log_norm,
            id_=id_
        )

    def value_for(self, unit: float) -> float:
        """
        Map a unit interval value (0 to 1) to a value consistent with the Beta distribution.

        Parameters
        ----------
        unit
            Input value in the unit interval [0, 1].

        Returns
        -------
        float
            Corresponding Beta-distributed value.

        Raises
        ------
        NotImplementedError
            This method should be implemented by subclasses.
        """
        raise NotImplemented()

    def log_partition(self, xp=np) -> np.ndarray:
        """
        Compute the log partition function (log normalization constant) of the Beta distribution.

        Returns
        -------
        The value of the log Beta function, i.e. betaln(alpha, beta).
        """
        from scipy.special import betaln

        return betaln(*self.parameters)

    def natural_parameters(self, xp=np) -> np.ndarray:
        """
        Compute the natural parameters of the Beta distribution.

        Returns
        -------
        np.ndarray
            Natural parameters array [alpha - 1, beta - 1].
        """
        return self.calc_natural_parameters(
            self.alpha,
            self.beta,
            xp=xp
        )

    @staticmethod
    def calc_natural_parameters(
            alpha: Union[float, np.ndarray],
            beta: Union[float, np.ndarray],
            xp=np
    ) -> np.ndarray:
        """
        Calculate the natural parameters of a Beta distribution from alpha and beta.

        Parameters
        ----------
        alpha
            Alpha (shape) parameter(s) of the Beta distribution.
        beta
            Beta (shape) parameter(s) of the Beta distribution.

        Returns
        -------
        Natural parameters [alpha - 1, beta - 1].
        """
        return xp.array([alpha - 1, beta - 1])

    @staticmethod
    def invert_natural_parameters(
            natural_parameters: np.ndarray
    ) -> np.ndarray:
        """
        Convert natural parameters back to standard Beta distribution parameters.

        Parameters
        ----------
        natural_parameters
            Array of natural parameters [alpha - 1, beta - 1].

        Returns
        -------
        Standard Beta parameters [alpha, beta].
        """
        return natural_parameters + 1

    @classmethod
    def invert_sufficient_statistics(
            cls, sufficient_statistics: np.ndarray
    ) -> np.ndarray:
        """
        Estimate natural parameters from sufficient statistics using inverse operations.

        Parameters
        ----------
        sufficient_statistics
            Sufficient statistics (e.g. expectations of log X and log(1 - X)).

        Returns
        -------
        Natural parameters computed from sufficient statistics.
        """
        a, b = inv_beta_suffstats(*sufficient_statistics)
        return cls.calc_natural_parameters(a, b)

    @classmethod
    def to_canonical_form(cls, x: np.ndarray, xp=np) -> np.ndarray:
        """
        Convert a value x in (0,1) to the canonical sufficient statistics for Beta.

        Parameters
        ----------
        x
            Values in the support of the Beta distribution (0 < x < 1).

        Returns
        -------
        Canonical sufficient statistics [log(x), log(1 - x)].
        """
        return xp.array([xp.log(x), xp.log1p(-x)])

    @cached_property
    def mean(self) -> Union[np.ndarray, float]:
        """
        Compute the mean of the Beta distribution.

        Returns
        -------
        Mean value alpha / (alpha + beta).
        """
        return self.alpha / (self.alpha + self.beta)

    @cached_property
    def variance(self) -> Union[np.ndarray, float]:
        """
        Compute the variance of the Beta distribution.

        Returns
        -------
        Variance value of the Beta distribution.
        """
        return (
                self.alpha * self.beta
                / (self.alpha + self.beta) ** 2
                / (self.alpha + self.beta + 1)
        )

    def sample(self, n_samples: int = None) -> np.ndarray:
        """
        Draw samples from the Beta distribution.

        Parameters
        ----------
        n_samples
            Number of samples to draw. If None, returns a single sample.

        Returns
        -------
        Samples drawn from Beta(alpha, beta).
        """
        a, b = self.parameters
        shape = (n_samples,) + self.shape if n_samples else self.shape
        return np.random.beta(a, b, size=shape)

    def kl(self, dist: "BetaMessage") -> float:
        """
        Calculate the Kullback-Leibler divergence KL(self || dist).

        Parameters
        ----------
        dist
            The Beta distribution to compare against.

        Returns
        -------
        float
            The KL divergence value.

        Raises
        ------
        TypeError
            If the support of the two distributions does not match.
        """
        from scipy.special import betaln
        from scipy.special import psi

        # TODO check this is correct
        # https://arxiv.org/pdf/0911.4863.pdf
        if self._support != dist._support:
            raise TypeError('Support does not match')

        aP, bP = dist.parameters
        aQ, bQ = self.parameters
        return (
                betaln(aQ, bQ) - betaln(aP, bP)
                - (aQ - aP) * psi(aP)
                - (bQ - bP) * psi(bP)
                + (aQ - aP + bQ - bP) * psi(aP + bP)
        )

    def logpdf_gradient(
            self,
            x: Union[float, np.ndarray]
    ) -> Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]:
        """
        Compute the log probability density function and its gradient at x.

        Parameters
        ----------
        x
            Point(s) in (0, 1) where to evaluate the logpdf and gradient.

        Returns
        -------
        logl
            Log of the PDF evaluated at x.
        gradl
            Gradient of the log PDF at x.
        """
        logl = self.logpdf(x)
        a, b = self.parameters
        gradl = (a - 1) / x + (b - 1) / (x - 1)
        return logl, gradl

    def logpdf_gradient_hessian(
            self,
            x: Union[float, np.ndarray]
    ) -> Tuple[Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray]]:
        """
        Compute the logpdf, its gradient, and Hessian at x.

        Parameters
        ----------
        x
            Point(s) in (0, 1) where to evaluate the logpdf, gradient, and Hessian.

        Returns
        -------
        logl
            Log of the PDF evaluated at x.
        gradl
            Gradient of the log PDF at x.
        hessl
            Hessian (second derivative) of the log PDF at x.
        """
        logl = self.logpdf(x)
        a, b = self.parameters
        ax, bx = (a - 1) / x, (b - 1) / (x - 1)
        gradl = ax + bx
        hessl = -ax / x - bx / (x - 1)
        return logl, gradl, hessl



import logging
import os
import re
import shutil
import zipfile
from abc import ABC, abstractmethod
from configparser import NoSectionError
from os import path
from pathlib import Path
from typing import Optional

import numpy as np

from autoconf import conf
from autofit.mapper.identifier import Identifier, IdentifierField
from autofit.non_linear.samples.summary import SamplesSummary

from autofit.text import text_util
from autofit.tools.util import open_, zip_directory

logger = logging.getLogger(__name__)

pattern = re.compile(r"(?<!^)(?=[A-Z])")


class AbstractPaths(ABC):
    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[os.PathLike] = None,
        is_identifier_in_paths=True,
        parent: Optional["AbstractPaths"] = None,
        unique_tag: Optional[str] = None,
        identifier: str = None,
        image_path_suffix: str = "",
    ):
        """
        Manages the path structure for `NonLinearSearch` output, for analyses both not using and using the search
        API. Use via non-linear searches requires manual input of paths, whereas the search API manages this using the
        search attributes.

        The output path within which the *Paths* objects path structure is contained is set via PyAutoConf, using the
        command:

        from autoconf import conf
        conf.instance = conf.Config(output_path="path/to/output")

        If we assume all the input strings above are used with the following example names:

        name = "name"
        path_prefix = "folder_0/folder_1"

        The output path of the `NonLinearSearch` results will be:

        /path/to/output/folder_0/folder_1/name

        Parameters
        ----------
        name
            The name of the non-linear search, which is used as a folder name after the ``path_prefix``. For searchs
            this name is the ``name``.
        path_prefix
            A prefixed path that appears after the output_path but before the name variable.
        is_identifier_in_paths
            If True output path and symlink path terminate with an identifier generated from the
            search and model
        parent
            The parent paths object of this paths object.
        unique_tag
            A unique tag for the search, used to differentiate between searches with the same name.
        identifier
            A custom identifier for the search, if this is not None it will be used instead of the automatically
            generated identifier
        image_path_suffix
            A suffix which is appended to the image path. This is used to differentiate between different
            image outputs, for example the image of the starting point of an MLE.
        """

        self.name = name or ""
        self.path_prefix = path_prefix or ""

        self.unique_tag = unique_tag

        self._non_linear_name = None
        self.__custom_identifier = identifier
        self.__identifier = None

        self.is_identifier_in_paths = is_identifier_in_paths

        self._parent = None
        self.parent = parent

        try:
            self.remove_files = conf.instance["general"]["output"]["remove_files"]

            if conf.instance["general"]["hpc"]["hpc_mode"]:
                self.remove_files = True
        except NoSectionError as e:
            logger.exception(e)

        self.image_path_suffix = image_path_suffix

    @property
    @abstractmethod
    def samples(self):
        pass

    def save_parent_identifier(self):
        pass

    def save_unique_tag(self, is_grid_search=False):
        pass

    def __str__(self):
        return str(self.output_path)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    @property
    def parent(self) -> "AbstractPaths":
        """
        The search performed before this search. For example, a search
        that is then compared to searches during a grid search.
        """
        return self._parent

    @parent.setter
    @abstractmethod
    def parent(self, parent: "AbstractPaths"):
        pass

    @property
    @abstractmethod
    def is_grid_search(self) -> bool:
        pass

    def for_sub_analysis(self, analysis_name: str):
        return self.create_child(name=analysis_name)

    @abstractmethod
    def create_child(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths: Optional[bool] = None,
        identifier: Optional[str] = None,
    ) -> "AbstractPaths":
        """
        Create a paths object which is the child of some parent
        paths object. This is done during a GridSearch so that
        results can be stored in the correct directory.

        Parameters
        ----------
        name
        path_prefix
        is_identifier_in_paths
            If False then this path's identifier will not be
            added to its output path.
        identifier

        Returns
        -------
        A new paths object
        """

    search = IdentifierField()
    model = IdentifierField()
    unique_tag = IdentifierField()

    @property
    def non_linear_name(self):
        if self._non_linear_name is None:
            if self.search is not None:
                self._non_linear_name = pattern.sub(
                    "_", type(self.search).__name__
                ).lower()
        return self._non_linear_name

    @property
    def _identifier(self):
        if self.__custom_identifier is not None:
            return self.__custom_identifier
        if self.__identifier is None:
            if None in (self.model, self.search):
                logger.debug(
                    "Generating identifier without both model and search having been set."
                )

            identifier_list = [self.search, self.model]

            if self.unique_tag is not None:
                identifier_list.append(self.unique_tag)
            self.__identifier = Identifier(identifier_list)

        return self.__identifier

    @_identifier.setter
    def _identifier(self, identifier):
        self.__identifier = identifier

    @property
    def identifier(self):
        return str(self._identifier)

    def save_identifier(self):
        with open_(self.output_path / ".identifier", "w+") as f:
            f.write(self._identifier.description)

    @property
    def search_internal_path(self) -> Path:
        """
        The path to the samples folder.
        """

        os.makedirs(self._files_path / "search_internal", exist_ok=True)

        return self._files_path / "search_internal"

    @property
    def image_path(self) -> Path:
        """
        The path to the image folder.
        """

        if not os.path.exists(self.output_path / f"image{self.image_path_suffix}"):
            os.makedirs(self.output_path / f"image{self.image_path_suffix}")

        return self.output_path / f"image{self.image_path_suffix}"

    @property
    def profile_path(self) -> Path:
        """
        The path to the profile folder.
        """
        return self.output_path / "profile"

    @property
    def output_path(self) -> Path:
        """
        The path to the output information for a search.
        """

        strings = list(
            filter(
                None,
                [
                    str(conf.instance.output_path),
                    str(self.path_prefix),
                    self.unique_tag,
                    str(self.name),
                ],
            )
        )

        if self.is_identifier_in_paths:
            strings.append(self.identifier)

        return Path(path.join("", *strings))

    @property
    def _files_path(self) -> Path:
        """
        This is private for a reason, use the save_json etc. methods to save and load json
        """
        files_path = self.output_path / "files"
        try:
            os.makedirs(files_path, exist_ok=True)
        except FileExistsError:
            pass
        return files_path

    def zip_remove(self):
        """
        Copy files from the sym linked search folder then remove the sym linked folder.
        """

        self._zip()

    def _zip(self):
        try:
            zip_directory(self.output_path, self._zip_path)

            if self.remove_files:
                shutil.rmtree(
                    self.output_path,
                    ignore_errors=True,
                )

        except FileNotFoundError:
            pass

    def restore(self):
        """
        Copy files from the ``.zip`` file to the samples folder.
        """

        if path.exists(self._zip_path):
            shutil.rmtree(self.output_path, ignore_errors=True)

            try:
                try:
                    with zipfile.ZipFile(self._zip_path, "r") as f:
                        f.extractall(self.output_path)
                except FileExistsError:
                    pass
            except zipfile.BadZipFile as e:
                raise zipfile.BadZipFile(
                    f"Unable to restore the zip file at the path {self._zip_path}"
                ) from e

            try:
                os.remove(self._zip_path)
            except FileNotFoundError:
                pass

    def __eq__(self, other):
        return isinstance(other, AbstractPaths) and all(
            [
                self.path_prefix == other.path_prefix,
                self.name == other.name,
                self.non_linear_name == other.non_linear_name,
            ]
        )

    @property
    def _zip_path(self) -> str:
        return f"{self.output_path}.zip"

    @abstractmethod
    def save_json(self, name, object_dict: dict, prefix: str = ""):
        pass

    @abstractmethod
    def load_json(self, name, prefix: str = "") -> dict:
        pass

    @abstractmethod
    def save_array(self, name, array: np.ndarray):
        pass

    @abstractmethod
    def load_array(self, name) -> np.ndarray:
        pass

    @abstractmethod
    def save_fits(self, name: str, fits, prefix: str = ""):
        pass

    @abstractmethod
    def load_fits(self, name: str, prefix: str = ""):
        pass

    @abstractmethod
    def save_object(self, name: str, obj: object, prefix: str = ""):
        pass

    @abstractmethod
    def load_object(self, name: str, prefix: str = ""):
        pass

    @abstractmethod
    def remove_object(self, name: str):
        pass

    @abstractmethod
    def is_object(self, name: str) -> bool:
        pass

    def save_search_internal(self, obj):
        raise NotImplementedError

    def load_search_internal(self):
        raise NotImplementedError

    def remove_search_internal(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def is_complete(self) -> bool:
        pass

    @abstractmethod
    def completed(self):
        pass

    @abstractmethod
    def save_all(self, search_config_dict=None, info=None):
        pass

    @abstractmethod
    def load_samples(self):
        """
        Load samples from the database
        """

    @abstractmethod
    def save_samples(self, samples):
        """
        Save samples to the database
        """

    def save_samples_summary(
        self,
        samples_summary: SamplesSummary,
        name: str = "",
    ):
        """
        Save samples summary to the database.
        """

    def load_samples_summary(self) -> SamplesSummary:
        """
        Load samples summary from the database.
        """

    @abstractmethod
    def save_latent_samples(self, latent_samples):
        """
        Save latent variables. These are values computed from an instance and output
        during analysis.
        """

    @abstractmethod
    def load_samples_info(self):
        pass

    def save_summary(
        self,
        samples,
        latent_samples,
        log_likelihood_function_time,
        visualization_time = None,
    ):
        result_info = text_util.result_info_from(
            samples=samples,
        )
        self.output_model_results(result_info=result_info)

        if latent_samples:
            result_info = text_util.result_info_from(
                samples=latent_samples,
            )
            filename = self.output_path / "latent.results"
            with open_(filename, "w") as f:
                f.write(result_info)

        text_util.search_summary_to_file(
            samples=samples,
            log_likelihood_function_time=log_likelihood_function_time,
            visualization_time=visualization_time,
            filename=self.output_path / "search.summary",
        )

    @property
    def _samples_file(self) -> Path:
        return self._files_path / "samples.csv"

    @property
    def _latent_variables_file(self) -> Path:
        return self._files_path / "latent.csv"

    @property
    def _covariance_file(self) -> Path:
        return self._files_path / "covariance.csv"

    @property
    def _info_file(self) -> Path:
        return self._files_path / "samples_info.json"

    def output_model_results(self, result_info):

        filename = self.output_path / "model.results"

        with open_(filename, "w") as f:
            f.write(result_info)


from typing import Optional

import numpy as np

from autofit.messages.normal import NormalMessage
from .abstract import Prior
from ...messages.composed_transform import TransformedMessage
from ...messages.transform import log_transform


class LogGaussianPrior(Prior):
    __identifier_fields__ = ("mean", "sigma")
    __database_args__ = ("mean", "sigma", "id_")

    def __init__(
        self,
        mean: float,
        sigma: float,
        id_: Optional[int] = None,
    ):
        """
        A prior for a variable whose logarithm is gaussian distributed. Work in natural log.

        The conversion of an input unit value, ``u``, to a physical value, ``p``, via the prior is as follows:

        .. math::

            p = \mu + (\sigma * sqrt(2) * erfcinv(2.0 * (1.0 - u))

        For example for ``prior = LogGaussianPrior(mean=1.0, sigma=2.0)``, an
        input ``prior.value_for(unit=0.5)`` is equal to 1.0.

        [Rich describe how this is done via message]

        Parameters
        ----------
        mean
            The *natural log* of the distribution's mean.
        sigma
            The spread of this distribution in *natural log* space, e.g. sigma=1.0 means P(ln x) has a
            standard deviation of 1.

        Examples
        --------

        prior = af.LogGaussianPrior(mean=1.0, sigma=2.0)

        physical_value = prior.value_for(unit=0.5)
        """

        self.mean = mean
        self.sigma = sigma

        message = TransformedMessage(
            NormalMessage(mean, sigma),
            log_transform,
        )

        super().__init__(
            message=message,
            id_=id_,
        )

    def tree_flatten(self):
        return (
            self.mean,
            self.sigma,
            self.id,
        ), ()

    @classmethod
    def with_limits(cls, lower_limit: float, upper_limit: float) -> "LogGaussianPrior":
        """
        Create a new gaussian prior centred between two limits
        with sigma distance between this limits.

        Note that these limits are not strict so exceptions will not
        be raised for values outside of the limits.

        This function is typically used in prior passing, where the
        result of a model-fit are used to create new Gaussian priors
        centred on the previously estimated median PDF model.

        Parameters
        ----------
        lower_limit
            The lower limit of the new Gaussian prior.
        upper_limit
            The upper limit of the new Gaussian Prior.

        Returns
        -------
        A new GaussianPrior
        """
        return cls(
            mean=(lower_limit + upper_limit) / 2,
            sigma=upper_limit - lower_limit,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
        )

    def _new_for_base_message(self, message):
        """
        Create a new instance of this wrapper but change the parameters used
        to instantiate the underlying message. This is useful for retaining
        the same transform stack after recreating the underlying message during
        projection.
        """
        return LogGaussianPrior(
            *message.parameters,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
            id_=self.instance().id,
        )

    def value_for(self, unit: float) -> float:
        """
        Return a physical value for a value between 0 and 1 with the transformation
        described by this prior.

        Parameters
        ----------
        unit
            A unit value between 0 and 1.

        Returns
        -------
        A physical value, mapped from the unit value accoridng to the prior.
        """
        return super().value_for(unit)

    @property
    def parameter_string(self) -> str:
        return f"mean = {self.mean}, sigma = {self.sigma}"

    def log_prior_from_value(self, value, xp=np):
        if value <= 0:
            return float("-inf")

        return self.message.base_message.log_prior_from_value(np.log(value)) - np.log(
            value
        )



import os

from .sqlalchemy_ import sa

from autoconf import conf
from .aggregator import *
from .migration.steps import migrator
from .model import *


def open_database(
        filename="database.sqlite"
) -> sa.orm.Session:
    """
    Open a database file in the output directory or connect to an
    existing database.

    If filename ends with ".sqlite" then an existing database file
    is opened or a new one is created.

    Otherwise, filename is assumed to describe the URL for an existing
    database which is connected to.

    To connect to a postgres database one must be created.

    1) Install postgres https://www.postgresql.org/download/
    2) Install the python postgres client ``psycopg2==2.9.1``
    2) Create a user ``createuser autofit``
    3) Create a database ``createdb -O autofit autofit``
    4) Open that database using this function

       open_database(
           "postgresql://autofit@localhost/autofit"
       )

    Note that the above instructions create a database called autofit
    with a user called autofit. You can create a new database for the
    same user:

        createdb -O autofit new_database

    Also note that the user has no password. This is fine for developing
    locally but if you ever find yourself making a database that other
    people might be able to access you might want to give the user a
    password.

    Parameters
    ----------
    filename
        The name for the database file including sqlite suffix

    Returns
    -------
    A SQLAlchemy session
    """
    if filename.endswith(
            ".sqlite"
    ):
        output_path = conf.instance.output_path

        if not filename.startswith("/"):
            filename = f"{output_path}/{filename}"

        os.makedirs(
            "/".join(filename.split("/")[:-1]),
            exist_ok=True
        )

        exists = os.path.exists(filename)

        string = f'sqlite:///{filename}'
        kwargs = dict(
            connect_args={'timeout': 15}
        )

    else:
        string = filename
        kwargs = dict()
        exists = True

    engine = sa.create_engine(
        string,
        **kwargs
    )
    session = sa.orm.sessionmaker(bind=engine)()
    if exists:
        migrator.migrate(
            session
        )
    else:
        Base.metadata.create_all(engine)
    return session



import numpy as np



def psilog(x: np.ndarray) -> np.ndarray:
    """
    psi(x) - log(x)
    needed when calculating E[ln[x]] when x is a Gamma variable
    """
    from scipy import special

    return special.digamma(x) - np.log(x)


def grad_psilog(x: np.ndarray) -> np.ndarray:
    """d_x (psi(x) - log(x)) = psi^1(x) - 1/x

    needed when calculating the inverse of psilog(x)
    by using Newton-Raphson

    see:
    invpsilog(c)
    """
    from scipy import special

    return special.polygamma(1, x) - 1 / x


def invpsilog(c: np.ndarray) -> np.ndarray:
    """
    Solves the equation

    psi(x) - log(x) = c

    where psi is the digamma function. c must be negative.
    The function calculates an approximate inverse which it uses as
    a starting point to 4 iterations of the Newton-Raphson algorithm.
    """
    c = np.asanyarray(c)

    if not np.all(c < 0):
        raise ValueError("values passed must be negative")

    # approximate starting guess
    # -1/x < psilog(x) < -1/(2x)
    A, beta, gamma = 0.38648347, 0.89486989, 0.78578843
    x0 = -(1 - 0.5 * (1 + A * (-c) ** beta) ** -gamma) / c

    # do 4 iterations of Newton Raphson to refine estimate
    for _ in range(4):
        f0 = psilog(x0) - c
        x0 = x0 - f0 / grad_psilog(x0)

    return x0



import shutil

import dill
import json
import numpy as np
import os
from os import path
from pathlib import Path
from typing import Optional, Union, cast, Type
import logging

from autoconf import conf
from autoconf.class_path import get_class
from autoconf.dictable import to_dict, from_dict
from autoconf.output import conditional_output, should_output
from autofit.text import formatter
from autofit.tools.util import open_
from autofit.non_linear.samples.samples import Samples

from .abstract import AbstractPaths

from ..samples import load_from_table
from autofit.non_linear.samples.pdf import SamplesPDF
from autofit.non_linear.samples.summary import SamplesSummary

from ...visualise import VisualiseGraph

logger = logging.getLogger(__name__)


class DirectoryPaths(AbstractPaths):
    def _path_for_pickle(self, name: str, prefix: str = "") -> Path:
        return self._files_path / prefix / f"{name}.pickle"

    def _path_for_json(self, name, prefix: str = "") -> Path:
        if isinstance(name, Path):
            return name
        return self._files_path / prefix / f"{name}.json"

    def _path_for_csv(self, name) -> Path:
        return self._files_path / f"{name}.csv"

    def _path_for_fits(self, name, prefix: str = "") -> Path:
        os.makedirs(self._files_path / prefix, exist_ok=True)

        return self._files_path / prefix / f"{name}.fits"

    @conditional_output
    def save_object(self, name: str, obj: object, prefix: str = ""):
        """
        Serialise an object using dill and save it to the pickles
        directory of the search.

        Parameters
        ----------
        name
            The name of the object
        obj
            A serialisable object
        prefix
            A prefix to add to the path which is the name of the folder the file is saved in.
        """
        with open_(self._path_for_pickle(name, prefix), "wb") as f:
            dill.dump(obj, f)

    @conditional_output
    def save_json(self, name, object_dict: Union[dict, list], prefix: str = ""):
        """
        Save a dictionary as a json file in the jsons directory of the search.

        Parameters
        ----------
        name
            The name of the json file
        object_dict
            The dictionary to save
        prefix
            A prefix to add to the path which is the name of the folder the file is saved in.
        """
        with open_(self._path_for_json(name, prefix), "w+") as f:
            json.dump(object_dict, f, indent=4)

    def load_json(self, name, prefix: str = ""):
        with open_(self._path_for_json(name, prefix)) as f:
            return json.load(f)

    @conditional_output
    def save_array(self, name: str, array: np.ndarray):
        """
        Save a numpy array as a csv file in the csvs directory of the search.

        Parameters
        ----------
        name
            The name of the csv file
        array
            The numpy array to save
        """
        # noinspection PyTypeChecker
        np.savetxt(self._path_for_csv(name), array, delimiter=",")

    def load_array(self, name: str):
        return np.loadtxt(self._path_for_csv(name), delimiter=",")

    @conditional_output
    def save_fits(self, name: str, fits, prefix: str = ""):
        """
        Save an HDU as a fits file in the fits directory of the search.

        Parameters
        ----------
        name
            The name of the fits file
        fits
            The HDUList to save
        prefix
            A prefix to add to the path which is the name of the folder the file is saved in.
        """
        fits.writeto(self._path_for_fits(name, prefix), overwrite=True)

    def load_fits(self, name: str, prefix: str = ""):
        """
        Load an HDU from a fits file in the fits directory of the search.

        Parameters
        ----------
        name
            The name of the fits file
        prefix
            A prefix to add to the path which is the name of the folder the file is saved in.

        Returns
        -------
        The loaded HDU.
        """
        from astropy.io import fits

        return fits.open(self._path_for_fits(name, prefix))[0]

    def load_object(self, name: str, prefix: str = ""):
        """
        Load a serialised object with the given name.

        e.g. if the name is 'model' then pickles/model.pickle is loaded.

        Parameters
        ----------
        name
            The name of a serialised object
        prefix
            A prefix to add to the path which is the name of the folder the file is saved in.

        Returns
        -------
        The deserialised object
        """
        with open_(self._path_for_pickle(name, prefix), "rb") as f:
            return dill.load(f)

    def remove_object(self, name: str):
        """
        Remove the object with the given name from the pickles folder.

        Parameters
        ----------
        name
            The name of a pickle file excluding .pickle
        """
        try:
            os.remove(self._path_for_pickle(name))
        except FileNotFoundError:
            pass

    def is_object(self, name: str) -> bool:
        """
        Is there a file pickles/{name}.pickle?
        """
        return os.path.exists(self._path_for_pickle(name))

    @property
    def is_complete(self) -> bool:
        """
        Has the search been completed?
        """
        return path.exists(self._has_completed_path)

    def save_search_internal(self, obj):
        """
        Save the internal representation of a non-linear search as dill file.

        The results in this representation are required to use a search's in-built tools for visualization,
        analysing samples and other tasks.
        """
        filename = self.search_internal_path / "search_internal.dill"

        with open_(filename, "wb") as f:
            dill.dump(obj, f)

    def load_search_internal(self):
        """
        Load the internal representation of a non-linear search from a pickle or dill file.

        The results in this representation are required to use a search's in-built tools for visualization,
        analysing samples and other tasks.

        Returns
        -------
        The results of the non-linear search in its internal representation.
        """

        # This is a nasty hack to load emcee backends. It will be removed once the source code is more stable.

        try:
            import emcee

            backend_filename = self.search_internal_path / "search_internal.hdf"
            if os.path.isfile(backend_filename):
                return emcee.backends.HDFBackend(filename=str(backend_filename))
        except ImportError:
            pass

        filename = self.search_internal_path / "search_internal.dill"

        with open_(filename, "rb") as f:
            return dill.load(f)

    def remove_search_internal(self):
        """
        Remove the internal representation of a non-linear search.

        This deletes the entire `search_internal` folder, including a .pickle / .dill file containing the interal
        results and files with the timer values.

        This folder can often have a large filesize, thus deleting it can reduce hard-disk use of the model-fit.
        """
        shutil.rmtree(self.search_internal_path)

    def completed(self):
        """
        Mark the search as complete by saving a file
        """
        open_(self._has_completed_path, "w+").close()

    def load_samples(self):
        return load_from_table(filename=self._samples_file)

    @property
    def samples(self):
        """
        Load the samples associated with the search from the output directory.
        """
        sample_list = self.load_samples()
        samples_info = self.load_samples_info()

        cls = cast(Type[Samples], get_class(samples_info["class_path"]))

        return cls.from_list_info_and_model(
            sample_list=sample_list,
            samples_info=samples_info,
            model=self.model,
        )

    def save_latent_samples(
        self,
        latent_samples,
    ):
        """
        Write out the latent variables of the model to a file.

        Parameters
        ----------
        latent_samples
            Samples describing the latent variables of the model
        """
        self._save_samples(latent_samples, name="latent")

    def save_samples(self, samples):
        """
        Save the final-result samples associated with the phase as a pickle
        """
        self._save_samples(samples)

    def _save_samples(self, samples, name=None):
        """
        Save the final-result samples associated with the phase as a pickle
        """

        if name is not None:
            directory = self._files_path / name
        else:
            directory = self._files_path
            name = "samples"
        if conf.instance["general"]["output"]["samples_to_csv"] and should_output(name):
            self.save_json(directory / "samples_info.json", samples.samples_info)

            if isinstance(samples, SamplesPDF):
                try:
                    samples.save_covariance_matrix(directory / "covariance.csv")
                except (ValueError, ZeroDivisionError) as e:
                    logger.warning(
                        f"Could not save covariance matrix because of the following error:\n{e}"
                    )

            samples.write_table(filename=directory / "samples.csv")

    def save_samples_summary(
        self, samples_summary: SamplesSummary, name="samples_summary"
    ):
        model = samples_summary.model

        filter_args = tuple(
            name
            for name in (
                "errors_at_sigma_1",
                "errors_at_sigma_3",
                "values_at_sigma_1",
                "values_at_sigma_3",
                "max_log_likelihood_sample",
                "median_pdf_sample",
            )
            if not should_output(name)
        )

        samples_summary.model = None
        self.save_json(
            name,
            to_dict(
                samples_summary,
                filter_args=filter_args,
            ),
        )
        samples_summary.model = model

    def load_samples_summary(self) -> SamplesSummary:
        samples_summary = from_dict(self.load_json(name="samples_summary"))
        samples_summary.model = self.model

        return samples_summary

    def load_latent_samples(self):
        return load_from_table(filename=self._files_path / "latent/samples.csv")

    def load_samples_info(self):
        with open_(self._info_file) as infile:
            return json.load(infile)

    def save_all(self, search_config_dict=None, info=None):
        info = info or {}

        self.save_identifier()
        self.save_parent_identifier()
        self._save_model_info(model=self.model)
        # VisualiseGraph(
        #     model=self.model,
        # ).save(str(self.output_path / "model_graph.html"))

        if info:
            self.save_json("info", info)

        self.save_json("search", to_dict(self.search))
        try:
            info_start = self.search.initializer.info_from_model(model=self.model)
            self._save_model_start_point(info=info_start)
        except (NotImplementedError, AttributeError):
            pass

        self.save_json("model", to_dict(self.model))
        self._save_metadata(search_name=type(self.search).__name__.lower())

    @AbstractPaths.parent.setter
    def parent(self, parent: AbstractPaths):
        """
        The search performed before this search. For example, a search
        that is then compared to searches during a grid search.
        """
        self._parent = parent

    def save_parent_identifier(self):
        if self.parent is not None:
            with open_(self._parent_identifier_path, "w+") as f:
                f.write(self.parent.identifier)
            self.parent.save_unique_tag()

    def save_unique_tag(self, is_grid_search=False):
        if is_grid_search:
            with open_(self._grid_search_path, "w+") as f:
                if self.unique_tag is not None:
                    f.write(self.unique_tag)

    @property
    def _parent_identifier_path(self) -> Path:
        return self.output_path / ".parent_identifier"

    @property
    def _grid_search_path(self) -> Path:
        return self.output_path / ".is_grid_search"

    @property
    def is_grid_search(self) -> bool:
        """
        Is this a grid search which comprises a number of child searches?
        """
        return os.path.exists(self._grid_search_path)

    def create_child(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths: Optional[bool] = None,
        identifier: Optional[str] = None,
    ) -> "AbstractPaths":
        """
        Create a paths object which is the child of some parent
        paths object. This is done during a GridSearch so that
        results can be stored in the correct directory.

        Parameters
        ----------
        name
        path_prefix
        is_identifier_in_paths
            If False then this path's identifier will not be
            added to its output path.
        identifier

        Returns
        -------
        A new paths object
        """
        child = type(self)(
            name=name or self.name,
            path_prefix=path_prefix or self.path_prefix,
            is_identifier_in_paths=(
                is_identifier_in_paths
                if is_identifier_in_paths is not None
                else self.is_identifier_in_paths
            ),
            parent=self,
        )
        child.model = self.model
        child.search = self.search
        child._identifier = identifier
        return child

    def for_sub_analysis(self, analysis_name: str):
        """
        Paths for an analysis which is a child of another analysis.

        The analysis name forms a new directory on the end of the original
        analysis output path.
        """
        from .sub_directory_paths import SubDirectoryPaths

        return SubDirectoryPaths(parent=self, analysis_name=analysis_name)

    def _save_metadata(self, search_name):
        """
        Save metadata associated with the phase, such as the name of the pipeline, the
        name of the phase and the name of the dataset being fit
        """
        with open_(self.output_path / "metadata", "a") as f:
            f.write(
                f"""name={self.name}\nnon_linear_search={search_name}
            """
            )

    def _save_model_info(self, model):
        """
        Save the model.info file, which summarises every parameter and prior.
        """
        with open_(self.output_path / "model.info", "w+") as f:
            f.write(model.info)

        try:
            with open_(self.output_path / "model.graph", "w+") as f:
                f.write( model.graph_info)
        except AttributeError:
            pass

    def _save_model_start_point(self, info):
        """
        Save the model.start file, which summarises the start point of every parameter.
        """
        with open_(self.output_path / "model.start", "w+") as f:
            f.write(info)

    def _save_parameter_names_file(self, model):
        """
        Create the param_names file listing every parameter's label and Latex tag, which is used for corner.py
        visualization.

        The parameter labels are determined using the label.ini and label_format.ini config files.
        """

        parameter_names = model.model_component_and_parameter_names
        parameter_labels = model.parameter_labels
        subscripts = model.superscripts_overwrite_via_config
        parameter_labels_with_subscript = [
            f"{label}_{subscript}"
            for label, subscript in zip(parameter_labels, subscripts)
        ]

        parameter_name_and_label = []

        for i in range(model.prior_count):
            line = formatter.add_whitespace(
                str0=parameter_names[i],
                str1=parameter_labels_with_subscript[i],
                whitespace=70,
            )
            parameter_name_and_label += [f"{line}\n"]

        formatter.output_list_of_strings_to_file(
            file=self._files_path / "model.paramnames",
            list_of_strings=parameter_name_and_label,
        )

    @property
    def _info_file(self) -> Path:
        return self._files_path / "samples_info.json"

    @property
    def _has_completed_path(self) -> Path:
        """
        A file indicating that a `NonLinearSearch` has been completed previously
        """
        return self.output_path / ".completed"

    def _make_path(self) -> str:
        """
        Returns the path to the folder at which the metadata should be saved

        The path terminates with the identifier, unless the identifier has already
        been added to the path.
        """
        path_ = Path(path.join(conf.instance.output_path, self.path_prefix, self.name))
        if self.is_identifier_in_paths:
            path_ = path_ / self.identifier
        return path_



from autofit.non_linear.analysis import Analysis
from autofit.non_linear.samples import Sample

def samples_with_log_likelihood_list(
        log_likelihood_list
):
    return [
        Sample(
            log_likelihood=log_likelihood,
            log_prior=0,
            weight=0
        )
        for log_likelihood
        in log_likelihood_list
    ]


class MockAnalysis(Analysis):
    prior_count = 2

    def __init__(self):
        super().__init__()
        self.fit_instances = list()

    def log_likelihood_function(self, instance):
        self.fit_instances.append(instance)
        return [1]

    def visualize(self, paths, instance, during_analysis):
        pass

    def log(self, instance):
        pass




import math
import pathlib
import warnings
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from autoconf import conf
from autoconf.output import should_output
from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples.sample import Sample, load_from_table
from autofit.non_linear.samples.samples import to_instance
from .samples import Samples
from .summary import SamplesSummary


class SamplesPDF(Samples):
    def __init__(
        self,
        model: AbstractPriorModel,
        sample_list: List[Sample],
        samples_info: Optional[Dict] = None,
    ):
        """
        Contains the samples of the non-linear search, including parameter values, log likelihoods,
        weights and other quantites.

        For example, the output class can be used to load an instance of the best-fit model, get an instance of any
        individual sample by the `NonLinearSearch` and return information on the likelihoods, errors, etc.

        This class stores samples of searches which provide the probability distribution function (PDF) of the
        model fit (e.g. nested samplers, MCMC).

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        sample_list
            The list of `Samples` which contains the paramoeters, likelihood, weights, etc. of every sample taken
            by the non-linear search.
        samples_info
            Contains information on the samples (e.g. total iterations, time to run the search, etc.).
        """

        super().__init__(
            model=model, sample_list=sample_list, samples_info=samples_info
        )

    def summary(self):
        median_pdf_sample = Sample.from_lists(
            model=self.model,
            parameter_lists=[self.median_pdf(as_instance=False)],
            log_likelihood_list=[self.max_log_likelihood_sample.log_likelihood],
            log_prior_list=[self.max_log_likelihood_sample.log_prior],
            weight_list=[self.max_log_likelihood_sample.weight],
        )[0]

        return SamplesSummary(
            model=self.model,
            max_log_likelihood_sample=self.max_log_likelihood_sample,
            median_pdf_sample=median_pdf_sample,
            log_evidence=self.log_evidence,
            errors_at_sigma_1=self.errors_at_sigma(
                sigma=1.0,
                as_dict=True,
            ),
            errors_at_sigma_3=self.errors_at_sigma(
                sigma=3.0,
                as_dict=True,
            ),
            values_at_sigma_1=self.values_at_sigma(
                sigma=1.0,
                as_dict=True,
            ),
            values_at_sigma_3=self.values_at_sigma(
                sigma=3.0,
                as_dict=True,
            ),
        )

    @classmethod
    def from_table(cls, filename: str, model):
        """
        Write a table of parameters, posteriors, priors and likelihoods

        Parameters
        ----------
        filename
            Where the table is to be written
        """
        from .stored import SamplesStored

        sample_list = load_from_table(filename=filename)

        return SamplesStored(model=model, sample_list=sample_list)

    @property
    def unconverged_sample_size(self):
        """
        If a set of samples are unconverged, alternative methods to compute their means, errors, etc are used.

        These use a subset of samples spanning the range from the most recent sample to the valaue of the
        unconverted_sample_size. However, if there are fewer samples than this size, we change the size to be the
        the size of the total number of samples.
        """

        unconverged_sample_size = conf.instance["general"]["output"][
            "unconverged_sample_size"
        ]

        if self.total_samples > unconverged_sample_size:
            return unconverged_sample_size
        return self.total_samples

    @property
    def pdf_converged(self) -> bool:
        """
        To analyse and visualize samples the analysis must be sufficiently converged to produce smooth enough
        PDF for error estimate and PDF generation.

        This property checks whether the non-linear search's samples are sufficiently converged for this, by checking
        if one sample's weight contains > 99% of the weight. If this is the case, it implies the convergence necessary
        for error estimate and visualization has not been met.

        This does not necessarily imply the `NonLinearSearch` has converged overall, only that errors and visualization
        can be performed numerically.
        """
        if np.max(self.weight_list) > 0.99:
            return False
        return True

    @to_instance
    def median_pdf(self) -> List[float]:
        """
        The median of the probability density function (PDF) of every parameter marginalized in 1D, returned
        as a model instance or list of values.
        """
        if self.pdf_converged:
            return [
                quantile(x=params, q=0.5, weights=self.weight_list)[0]
                for params in self.parameters_extract
            ]
        return self.max_log_likelihood(as_instance=False)

    @to_instance
    def values_at_sigma(self, sigma: float) -> [Tuple, ModelInstance]:
        """
        The value of every parameter marginalized in 1D at an input sigma value of its probability density function
        (PDF), returned as two lists of values corresponding to the lower and upper values parameter values.

        For example, if sigma is 1.0, the marginalized values of every parameter at 31.7% and 68.2% percentiles of each
        PDF is returned.

        This does not account for covariance between parameters. For example, if two parameters (x, y) are degenerate
        whereby x decreases as y gets larger to give the same PDF, this function will still return both at their
        upper values. Thus, caution is advised when using the function to reperform a model-fits.

        This is estimated using the `quantile` function if the samples have converged, by sampling the density
        function at an input PDF %. If not converged, a crude estimate using the range of values of the current
        physical live points is used.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """

        if self.pdf_converged:
            low_limit = (1 - math.erf(sigma / math.sqrt(2))) / 2

            lower_errors = [
                quantile(x=params, q=low_limit, weights=self.weight_list)[0]
                for params in self.parameters_extract
            ]
            upper_errors = [
                quantile(x=params, q=1 - low_limit, weights=self.weight_list)[0]
                for params in self.parameters_extract
            ]

            return [(lower, upper) for lower, upper in zip(lower_errors, upper_errors)]

        parameters_min = list(
            np.min(self.parameter_lists[-int(self.unconverged_sample_size) :], axis=0)
        )
        parameters_max = list(
            np.max(self.parameter_lists[-int(self.unconverged_sample_size) :], axis=0)
        )

        return [
            (parameters_min[index], parameters_max[index])
            for index in range(len(parameters_min))
        ]

    @to_instance
    def values_at_upper_sigma(self, sigma: float) -> Union[List, ModelInstance]:
        """
        The upper value of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a list.

        See values_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return list(
            map(lambda param: param[1], self.values_at_sigma(sigma, as_instance=False))
        )

    @to_instance
    def values_at_lower_sigma(self, sigma: float) -> Union[List, ModelInstance]:
        """
        The lower value of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a list.

        See values_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        ----------
        sigma
            The sigma limit within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return list(
            map(lambda param: param[0], self.values_at_sigma(sigma, as_instance=False))
        )

    @to_instance
    def errors_at_sigma(
        self, sigma: float, as_instance: bool = True
    ) -> [Tuple, ModelInstance]:
        """
        The lower and upper error of every parameter marginalized in 1D at an input sigma value of its probability
        density function (PDF), returned as a list.

        See values_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        error_vector_lower = self.errors_at_lower_sigma(sigma=sigma, as_instance=False)
        error_vector_upper = self.errors_at_upper_sigma(sigma=sigma, as_instance=False)
        return [
            (lower, upper)
            for lower, upper in zip(error_vector_lower, error_vector_upper)
        ]

    @to_instance
    def errors_at_upper_sigma(
        self, sigma: float, as_instance: bool = True
    ) -> Union[List, ModelInstance]:
        """
        The upper error of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a list.

        See values_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        uppers = self.values_at_upper_sigma(sigma=sigma, as_instance=False)
        return list(
            map(
                lambda upper, median_pdf: upper - median_pdf,
                uppers,
                self.median_pdf(as_instance=False),
            )
        )

    @to_instance
    def errors_at_lower_sigma(self, sigma: float) -> Union[List, ModelInstance]:
        """
        The lower error of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a list.

        See values_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        lowers = self.values_at_lower_sigma(sigma=sigma, as_instance=False)
        return list(
            map(
                lambda lower, median_pdf: median_pdf - lower,
                lowers,
                self.median_pdf(as_instance=False),
            )
        )

    @to_instance
    def error_magnitudes_at_sigma(self, sigma: float) -> Union[List, ModelInstance]:
        """
        The magnitude of every error after marginalization in 1D at an input sigma value of the probability density
        function (PDF), returned as two lists of values corresponding to the lower and upper errors.

        For example, if sigma is 1.0, the difference in the inferred values marginalized at 31.7% and 68.2% percentiles
        of each PDF is returned.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        uppers = self.values_at_upper_sigma(sigma=sigma, as_instance=False)
        lowers = self.values_at_lower_sigma(sigma=sigma, as_instance=False)
        return list(map(lambda upper, lower: upper - lower, uppers, lowers))

    @to_instance
    def draw_randomly_via_pdf(self) -> Union[List, ModelInstance]:
        """
        The parameter vector of an individual sample of the non-linear search drawn randomly from the PDF, returned as
        a 1D list.

        The draw is weighted by the sample weights to ensure that the sample is drawn from the PDF (which is important
        for non-linear searches like nested sampling).
        """

        sample_index = np.random.choice(
            a=range(len(self.sample_list)), p=self.weight_list
        )

        return self.parameter_lists[sample_index][:]

    def samples_drawn_randomly_via_pdf_from(self, total_draws: int = 100) -> "SamplesPDF":
        """
        Draw one or more samples randomly from the PDF, weighted by the sample weights.

        Parameters
        ----------
        total_draws : int, optional
            The number of samples to draw. Defaults to 100.

        Returns
        -------
        SamplesPDF
            A new SamplesPDF object containing the drawn samples.
        """
        # Normalize weights to sum to 1
        weights = np.asarray(self.weight_list, dtype=float)
        weights /= weights.sum()

        sample_indices = np.random.choice(
            a=len(self.sample_list),
            size=total_draws,
            replace=True,
            p=weights,
        )

        return SamplesPDF(
            model=self.model,
            sample_list=[self.sample_list[i] for i in sample_indices],
            samples_info=self.samples_info,
        )

    @to_instance
    def offset_values_via_input_values(
        self, input_vector: List
    ) -> Union[List, ModelInstance]:
        """
        The values of an input_vector offset by the median_pdf(as_instance=False) (the PDF medians).

        If the 'true' values of a model are known and input as the input_vector, this function returns the results
        of the `NonLinearSearch` as values offset from the 'true' model. For example, a value 0.0 means the non-linear
        search estimated the model parameter value correctly.

        Parameters
        ----------
        input_vector
            A 1D list of values the most-probable model is offset by. This list must have the same dimensions as free
            parameters in the model-fit.
        """
        return list(
            map(
                lambda input, median_pdf: median_pdf - input,
                input_vector,
                self.median_pdf(as_instance=False),
            )
        )

    @property
    def covariance_matrix(self) -> np.ndarray:
        """
        Compute the covariance matrix of the non-linear search samples, using the method `np.cov()` which is described
        at the following link:

        https://numpy.org/doc/stable/reference/generated/numpy.cov.html

        Follow that link for a description of what the covariance matrix is.

        This function may be called during a non-linear search, before the samples contain high likelihood regions
        that enable a robust covariance matrix to be computed. To reduce command line noise, the warning associated
        with this behaviour is suppressed.

        Returns
        -------
        ndarray
            A covariance matrix of shape [total_parameters, total_parameters] for the model parameters of the
            non-linear search.
        """
        if len(self.parameter_lists) == 1:
            return np.eye(1)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return np.cov(
                m=self.parameter_lists, rowvar=False, aweights=self.weight_list
            )

    def save_covariance_matrix(self, filename: Union[pathlib.Path, str]):
        """
        Save the covariance matrix as a CSV file.

        Parameters
        ----------
        filename
            The filename the covariance matrix is saved to.
        """
        if not should_output("covariance"):
            return
        # noinspection PyTypeChecker
        np.savetxt(filename, self.covariance_matrix, delimiter=",")

    @property
    def log_evidence(self):
        return None


def marginalize(
    parameter_list: List, sigma: float, weight_list: Optional[List] = None
) -> Tuple[float, float, float]:
    low_limit = (1 - math.erf(sigma / math.sqrt(2))) / 2

    median = quantile(x=parameter_list, q=0.5, weights=weight_list)[0]
    lower = quantile(x=parameter_list, q=low_limit, weights=weight_list)[0]
    upper = quantile(x=parameter_list, q=1 - low_limit, weights=weight_list)[0]

    return median, lower, upper


def quantile(x, q, weights=None):
    """
    Copied from corner.py

    Compute sample quantiles with support for weighted samples.

    Note
    ----
    When ``weight_list`` is ``None``, this method simply calls numpy's percentile
    function with the values of ``q`` multiplied by 100.

    Parameters
    ----------
    x : array_like[nsamples,]
       The samples.
    q : array_like[nquantiles,]
       The list of quantiles to compute. These should all be in the range
       ``[0, 1]``.
    weights: Optional[array_like[nsamples,]]
        An optional weight corresponding to each sample. These

    Returns
    -------
    quantiles : array_like[nquantiles,]
        The sample quantiles computed at ``q``.

    Raises
    ------
    ValueError
        For invalid quantiles; ``q`` not in ``[0, 1]`` or dimension mismatch between ``x`` and ``weight_list``.
    """
    x = np.atleast_1d(x)
    q = np.atleast_1d(q)

    if np.any(q < 0.0) or np.any(q > 1.0):
        raise ValueError("Quantiles must be between 0 and 1")

    if weights is None:
        return np.percentile(x, list(100.0 * q))
    else:
        weights = np.atleast_1d(weights)
        if len(x) != len(weights):
            raise ValueError("Dimension mismatch: len(weight_list) != len(x)")
        idx = np.argsort(x)
        sw = weights[idx]
        cdf = np.cumsum(sw)[:-1]
        cdf /= cdf[-1]
        cdf = np.append(0, cdf)
        return np.interp(q, cdf, x[idx]).tolist()



import logging
import multiprocessing
from abc import ABC, abstractmethod
from collections import abc
from itertools import count
from typing import Iterable

logger = logging.getLogger(
    __name__
)


class AbstractJobResult(ABC):
    def __init__(self, number):
        self.number = number

    def __eq__(self, other):
        return self.number == other.number

    def __gt__(self, other):
        return self.number > other.number

    def __lt__(self, other):
        return self.number < other.number


class AbstractJob(ABC):
    """
    Task to be completed in parallel
    """
    _number = count()

    def __init__(self, number=None):
        if number is None:
            number = next(self._number)
        self.number = number

    @abstractmethod
    def perform(self, *args):
        """
        Perform the task and return the result
        """


class Process(multiprocessing.Process):

    def __init__(
            self,
            name: str,
            job_queue: multiprocessing.Queue,
            initializer=None,
            initargs=None,
            job_args=tuple()
    ):
        """
        A parallel process that consumes Jobs through the job queue and outputs results through its own queue.

        Parameters
        ----------
        name: str
            The name of the process
        job_queue: multiprocessing.Queue
            The queue through which jobs are submitted
        """

        super().__init__(name=name)
        self.logger = logging.getLogger(
            f"process {name}"
        )

        self.logger.info("created")

        self.job_queue = job_queue
        self.queue = multiprocessing.Queue()

        self.initializer = initializer
        self.initargs = initargs

        self.job_args = job_args

    def _init(self):
        if self.initializer is not None:
            if self.initargs is None:
                return self.initializer()
            if isinstance(
                    self.initargs,
                    abc.Iterable
            ):
                initargs = tuple(self.initargs)
            else:
                initargs = (self.initargs,)

            self.initializer(
                *initargs
            )

    def run(self):
        """
        Run this process, completing each job in the job_queue and
        passing the result to the queue.
        """
        self._init()

        self.logger.debug("starting")
        while True:
            if self.job_queue.empty():
                break
            else:
                job = self.job_queue.get()
                try:
                    self.queue.put(
                        job.perform(
                            *self.job_args
                        )
                    )
                except Exception as e:
                    logger.exception(e)
                    self.queue.put(e)
        self.logger.debug("terminating")
        self.job_queue.close()

    @classmethod
    def run_jobs(
            cls,
            jobs: Iterable[AbstractJob],
            number_of_cores: int,
            initializer=None,
            initargs=None,
            job_args=tuple()
    ):
        """
        Run the collection of jobs across n - 1 other cores.

        Parameters
        ----------
        job_args
        initargs
        initializer
        jobs
            Serializable concrete children of the AbstractJob class
        number_of_cores
            The number of cores this computer has. Must be at least 2.
        """
        if number_of_cores < 2:
            raise AssertionError(
                "The number of cores available must be at least 2 for parallel to run"
            )

        job_queue = multiprocessing.Queue()

        processes = [
            cls(
                str(number),
                job_queue,
                initializer=initializer,
                initargs=initargs,
                job_args=job_args
            )
            for number in range(number_of_cores - 1)
        ]

        total = 0
        for job in jobs:
            job_queue.put(job)
            total += 1

        logger.info(
            f"Running {total} jobs across {number_of_cores} processes"
        )

        for process in processes:
            process.start()

        logger.debug("Starting processes")

        process_count = 0
        exception = None

        while process_count < total:
            for process in processes:
                while not process.queue.empty():
                    result = process.queue.get()
                    if isinstance(result, Exception):
                        process_count += 1
                        logger.exception(result)
                        exception = result
                    process_count += 1
                    yield result

        job_queue.close()

        logger.debug("Finished jobs")

        for process in processes:
            process.join(timeout=1.0)

        logger.debug("Joining processes")

        if exception is not None:
            raise AssertionError(
                exception
            )


class StopCommand:
    """
    A command that can be passed into a process queue to gracefully stop
    the process
    """



import warnings
from abc import ABC, abstractmethod
from functools import wraps
from typing import Tuple

import numpy as np

from ..mapper.operator import DiagonalMatrix, LinearOperator, ShermanMorrison

epsilon = 1e-14


def ndtri(x):

    from scipy.special import ndtri

    x = np.array(x, dtype=float)
    x[(x <= 0) & (x >= -epsilon)] = epsilon
    x[(x >= 1) & (x <= 1 + epsilon)] = 1 - epsilon
    return ndtri(x)


def numerical_jacobian(x, func, eps=1e-8, args=(), **kwargs):
    x = np.array(x)
    f0 = func(x, *args, **kwargs)
    jac = np.empty(np.shape(f0) + np.shape(x))
    fslice = (slice(None),) * np.ndim(f0)
    with np.nditer(x, flags=["multi_index"], op_flags=["readwrite"]) as it:
        for xi in it:
            xi += eps
            f1 = func(x, *args, **kwargs)
            jac[fslice + it.multi_index] = (f1 - f0) / eps
            xi -= eps

    return jac


class AbstractDensityTransform(ABC):
    """
    This class allows the transformation of a probability density function, p(x)
    whilst preserving the measure of the distribution, i.e.

    int p(x) dx = 1

    p'(f) = p(f(x)) * |df/dx|

    inf p'(f) df = 1

    Methods
    -------
    transform
        calculates f(x)

    inv_transform
        calculates f^{-1}(y)

    jacobian
        calculates df/dx

    log_det
        calculates log |df/dx|

    log_det_grad
        calculates |df/dx|, d log_det/dx

    transform_det
        calculates f(x), |df/dx|

    transform_jac
        calculates f(x), df/dx

    transform_det_jac
        calculates f(x), log_det, d log_det/dx, df/dx

    These final 3 functions are defined so that child classes
    can define custom methods that avoid recalculation of intermediate
    values that are needed to calculate multiple versions of the quantities
    """

    @abstractmethod
    def transform(self, x):
        pass

    @abstractmethod
    def inv_transform(self, x):
        pass

    @abstractmethod
    def jacobian(self, x: np.ndarray) -> LinearOperator:
        pass

    @abstractmethod
    def log_det(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def log_det_grad(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        pass

    def _numerical_jacobian_func(self, x):
        return self.log_det_grad(x)[1].sum(0)

    def log_det_hess(self, x: np.ndarray) -> np.ndarray:
        return numerical_jacobian(x, self._numerical_jacobian_func)

    def transform_det(self, x) -> Tuple[np.ndarray, np.ndarray]:
        return self.transform(x), self.log_det(x)

    def transform_jac(self, x) -> Tuple[np.ndarray, LinearOperator]:
        return self.transform(x), self.jacobian(x)

    def transform_det_jac(
        self, x
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, LinearOperator]:
        return (self.transform(x), *self.log_det_grad(x), self.jacobian(x))

    def transform_func(self, func):
        @wraps(func)
        def transformed_func(*args, **kwargs):
            x, *args = args
            x = self.transform(x)
            return func(x, *args, **kwargs)

        transformed_func.transform = self

        return transformed_func

    def transform_func_grad(self, func_grad):
        @wraps(func_grad)
        def transformed_func_grad(*args, **kwargs):
            x, *args = args
            x, jac = self.transform_jac(x)
            val, grad = func_grad(x, *args, **kwargs)
            return x, grad * jac

        transformed_func_grad.transform = self
        return transformed_func_grad

    def transform_func_grad_hess(self, func_grad_hess):
        @wraps(func_grad_hess)
        def transformed_func_grad_hess(*args, **kwargs):
            x, *args = args
            x, jac = self.transform_jac(x)
            val, grad, hess = func_grad_hess(x, *args, **kwargs)
            return val, grad * jac, jac.quad(hess)

        transformed_func_grad_hess.transform = self
        return transformed_func_grad_hess


class LinearTransform(AbstractDensityTransform):
    def __init__(self, linear: LinearOperator):
        self.linear = linear

    def transform(self, x: np.ndarray) -> np.ndarray:
        return self.linear * x

    def inv_transform(self, x: np.ndarray) -> np.ndarray:
        return self.linear.ldiv(x)

    def jacobian(self, x: np.ndarray) -> np.ndarray:
        return self.linear

    def log_det(self, x: np.ndarray) -> np.ndarray:
        return self.linear.log_det

    def log_det_grad(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        return self.log_det(x), 0


class LinearShiftTransform(LinearTransform):
    def __init__(self, shift: float = 0, scale: float = 1):
        self.shift = float(shift)
        self.scale = float(scale)
        super().__init__(DiagonalMatrix(np.reciprocal(self.scale)))

    def inv_transform(self, x: np.ndarray) -> np.ndarray:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return x * self.scale + self.shift

    def transform(self, x: np.ndarray) -> np.ndarray:
        return (x - self.shift) / self.scale

    def log_det(self, x: np.ndarray) -> np.ndarray:
        return -np.log(self.scale) * np.ones_like(x)


class FunctionTransform(AbstractDensityTransform):
    def __init__(self, func, inv_func, grad, hess=None, args=(), func_grad_hess=None):
        self.func = func
        self.inv_func = inv_func
        self.grad = grad
        self.hess = hess
        self.args = args
        self.func_grad_hess = func_grad_hess

    def transform(self, x):

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            return self.func(x, *self.args)

    def inv_transform(self, x):

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            return self.inv_func(x, *self.args)

    def jacobian(self, x):
        return DiagonalMatrix(self.grad(x, *self.args))

    def log_det(self, x: np.ndarray) -> np.ndarray:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            gs = self.grad(x, *self.args)
            return np.log(gs)

    def log_det_grad(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        if self.func_grad_hess:
            x0, gs, hs = self.func_grad_hess(x, *self.args)
        else:
            gs = self.grad(x, *self.args)
            hs = self.hess(x, *self.args)
        return np.log(gs), hs / gs

    def transform_det_jac(
        self, x
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, LinearOperator]:
        if self.func_grad_hess:
            x0, gs, hs = self.func_grad_hess(x, *self.args)
        else:
            x0 = self.func(x, *self.args)
            gs = self.grad(x, *self.args)
            hs = self.hess(x, *self.args)
        return x0, np.log(gs), hs / gs, DiagonalMatrix(gs)


def exp3(x):
    expx = np.exp(x)
    return (expx,) * 3


exp_transform = FunctionTransform(np.exp, np.log, np.exp, func_grad_hess=exp3)


def log3(x):
    ix = np.reciprocal(x)
    return np.log(x), ix, -np.square(ix)


log_transform = FunctionTransform(np.log, np.exp, np.reciprocal, func_grad_hess=log3)


def _log_10_inv(x):
    return 10**x

LOG10 = np.log(10)
def _log10_recip(x):
    return np.reciprocal(x) / LOG10

def log10_3(x):
    ix = _log10_recip(x)
    return np.log10(x), ix, -ix/x

# TODO: what should func_grad_hess look like? np.log10, ... ?
log_10_transform = FunctionTransform(np.log10, _log_10_inv, _log10_recip, func_grad_hess=log10_3)


def sigmoid(x, scale=1, shift=0):
    return scale / (1 + np.exp(-x)) + shift


def logit(x, scale=1, shift=0):
    x = (x - shift) / scale
    return np.log(x) - np.log1p(-x)


def sigmoid_grad(x, scale=1, shift=0):
    expx = np.exp(-x)
    return scale * expx / np.square(1 + expx)


def logit_grad(x, scale=1, shift=0):
    x = (x - shift) / scale
    return (np.reciprocal(x) + np.reciprocal(1 - x)) / scale


def logit_hess(x, scale=1, shift=0):
    x = (x - shift) / scale
    return np.reciprocal(1 - x) - np.reciprocal(x)


def logit_grad_hess(x, scale=1, shift=0):
    x = (x - shift) / scale
    ix = np.reciprocal(x)
    ix1 = np.reciprocal(1 - x)
    ix2 = np.square(ix)
    ix12 = np.square(ix1)
    return (np.log(x) - np.log1p(-x), (ix + ix1) / scale, (ix12 - ix2) / scale**2)


logistic_transform = FunctionTransform(
    logit, sigmoid, logit_grad, func_grad_hess=logit_grad_hess
)


def shifted_logistic(shift=0, scale=1):
    return FunctionTransform(
        logit, sigmoid, logit_grad, func_grad_hess=logit_grad_hess, args=(scale, shift)
    )


def ndtri_grad(x):
    from scipy.stats._continuous_distns import _norm_pdf

    return np.reciprocal(_norm_pdf(ndtri(x)))


def ndtri_grad_hess(x):
    from scipy.stats._continuous_distns import _norm_pdf

    f = ndtri(x)
    phi = _norm_pdf(f)
    grad = np.reciprocal(phi)
    hess = grad**2 * f
    return f, grad, hess

def lazy_ndtr(x):
    from scipy.special import ndtr
    return ndtr(x)

def lazy_ndtri(x):
    from scipy.special import ndtri
    return ndtri(x)

phi_transform = FunctionTransform(
    lazy_ndtri, lazy_ndtr, ndtri_grad, func_grad_hess=ndtri_grad_hess
)


class MultinomialLogitTransform(AbstractDensityTransform):
    """
    makes multinomial logististic transform from the p to x, where,

    x_i = log(p_i / (1 - sum(p)))
    p_i = exp(x_i) / (sum(exp(x_j) for x_j in x) + 1)

    When p's n-simplex is defined by,

    all(0 <= p_i <= 1 for p_i in p) and sum(p) < 1
    """

    def __init__(self, axis=-1):
        self.axis = axis

    def _validate(self, p):
        p = np.asanyarray(p)
        keepdims = np.ndim(p) == self.ndim + 1
        if not (keepdims or np.ndim(p) == self.ndim):
            raise ValueError(
                f"dimension of input must be {self.ndim} or {self.ndim + 1}"
            )

        return p, keepdims

    def transform(self, p):
        p = np.asanyarray(p)
        lnp1 = np.log(1 - np.sum(p, axis=self.axis, keepdims=True))
        lnp = np.log(p)
        return lnp - lnp1

    def inv_transform(self, x):
        expx = np.exp(x)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return expx / (expx.sum(axis=self.axis, keepdims=True) + 1)

    def jacobian(self, p):
        p = np.asanyarray(p)
        pn1 = 1 - np.sum(p, axis=-1, keepdims=True)
        # ln1p = np.log(pn1)
        # lnp = np.log(p)
        return ShermanMorrison(
            DiagonalMatrix(1 / p), 1 / np.sqrt(pn1) * np.ones_like(p)
        )

    def log_det(self, p):
        p = np.asanyarray(p)
        p1 = 1 - np.sum(p, axis=self.axis, keepdims=True)
        # Hack to make sure summation broadcasting works correctly
        log_d = (
            -np.log(p).sum(axis=self.axis, keepdims=True) - np.log(p1)
        ) * np.full_like(p, p1.size / p.size)
        return log_d

    def log_det_grad(self, p):
        p = np.asanyarray(p)
        p1 = 1 - np.sum(p, axis=self.axis, keepdims=True)
        # Hack to make sure summation broadcasting works correctly
        log_d = (
            -np.log(p).sum(axis=self.axis, keepdims=True) - np.log(p1)
        ) * np.full_like(p, p1.size / p.size)
        return log_d, 1 / p1 - 1 / p

    def transform_det_jac(
        self, p
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, LinearOperator]:
        p = np.asanyarray(p)
        pn1 = 1 - np.sum(p, axis=self.axis, keepdims=True)
        ln1p = np.log(pn1)
        lnp = np.log(p)
        x = lnp - ln1p
        # Hack to make sure summation broadcasting works correctly
        logd = (-lnp.sum(axis=self.axis, keepdims=True) - ln1p) * np.full_like(
            p, pn1.size / p.size
        )
        logd_grad = 1 / pn1 - 1 / p
        jac = ShermanMorrison(DiagonalMatrix(1 / p), 1 / np.sqrt(pn1) * np.ones_like(p))
        return (x, logd, logd_grad, jac)


multinomial_logit_transform = MultinomialLogitTransform()



from .abstract import AbstractMessage
from .beta import BetaMessage
from .fixed import FixedMessage
from .gamma import GammaMessage
from .normal import (
    NormalMessage,
    NaturalNormal, 
    UniformNormalMessage,
    LogNormalMessage,
    Log10NormalMessage,
    MultiLogitNormalMessage,
    Log10UniformNormalMessage,
)
